"grf" <-
  function(n, grid = "irreg", nx = round(sqrt(n)), ny = round(sqrt(n)), nsim = 1,
           xlims = c(0, 1), ylims = c(0, 1), cov.model = c("exponential", "matern",
                                               "gaussian", "spherical", "circular", "cubic", "wave",
                                               "powered.exponential", 
                                               "cauchy", "gneiting", "gneiting.matern", "pure.nugget"),
           cov.pars = stop("covariance parameters (sigmasq and phi) needed"), kappa = 0.5,
           nugget=0, lambda=1, method = c("cholesky", "svd", "eigen", "circular.embedding"),
           messages.screen = TRUE, ... )
{
  ##
  ## reading and checking input
  ##
  call <- match.call()
  method <- match.arg(method)
  cov.model <- match.arg(cov.model)
  if(!is.null(kappa))
    if(cov.model == "matern" & kappa == 0.5) cov.model <- "exponential"
  rseed <- .Random.seed
  tausq <- nugget
  if (is.vector(cov.pars)) {
    sigmasq <- cov.pars[1]
    phi <- cov.pars[2]
    nst <- 1
  }
  else {
    sigmasq <- cov.pars[, 1]
    phi <- cov.pars[, 2]
    nst <- nrow(cov.pars)
  }
  sill.tot <- tausq + sum(sigmasq)
  messa <- grf.aux1(nst, nugget, sigmasq, phi, kappa, cov.model)
  if (messages.screen) {
    cat(messa$nst) ; cat("\n")
    cat(messa$nugget); cat("\n")
    cat(messa$cov.structures); cat("\n")
    cat(paste("grf: decomposition algorithm used is: ", method, "\n"))
  }
  results <- list()
  ##
  ## defining the locations for the simulated data
  ##
  if (is.matrix(grid) | is.data.frame(grid)) {
    results$coords <- as.matrix(grid)
    if (messages.screen) 
      cat("grf: simulations on a grid provided by the user\n")
  }
  else {
    if (grid == "irreg") {
      results$coords <- cbind(x = runif(n, xlims[1], xlims[2]),
                              y = runif(n, ylims[1], ylims[2]))
      if (messages.screen) 
        cat(paste("grf: generating irregular grid with ", n, " points\n"))
    }
    else {
      results$coords <- as.matrix(expand.grid(x = seq(xlims[1], xlims[2], l = nx),
                                              y = seq(ylims[1], ylims[2], l = ny)))
      if (messages.screen) 
        cat(paste("grf: generating grid ", nx, " * ", ny, 
                  " with ", n, " points\n"))
    }
  }
  n <- nrow(results$coords)
  ##
  ## simulating data at locations defined by the matrix results$coords
  ##
  if (all(phi) == 0) {
    results$data <- matrix(rnorm((n * nsim), mean = 0, sd = sqrt(sill.tot)), 
                           nrow = n, ncol = nsim)
  }
  else {
    if (method == "circular.embedding") {
      if (grid == "irreg") 
        stop("grf: circular.embedding algorithm only allowed for regular grids")
      stepx <- (xlims[2] - xlims[1])/(nx - 1)
      stepy <- (ylims[2] - ylims[1])/(ny - 1)
      if (round(1e+08 * stepx) != round(1e+08 *stepy)) 
        stop("grf: distance between points must be the same in X and Y directions")
      temp <- list(n = n, nst = nst, sigmasq = sigmasq, 
                   xlims = xlims, ylims = ylims, stepx = stepx, 
                   cov.model = cov.model, phi = phi, kappa = kappa)
      results$data <- apply(as.matrix(1:nsim), 1, grf.aux3, temp = temp)
      if (nugget != 0) {
        results$data <- results$data + matrix(rnorm((n * nsim), sd = sqrt(nugget)), 
                                              ncol = nsim)
      }
    }
    else{
      results$data <- matrix(rnorm((n * nsim)), nrow = n, ncol = nsim)
      cov.decomp <- varcov.spatial(coords = results$coords, 
                                   cov.model = cov.model, kappa = kappa,
                                   nugget = nugget, cov.pars = cov.pars, 
                                   only.decomposition = TRUE,
                                   func.inv = method)
      results$data <- cov.decomp %*% results$data
    }
    if (nsim == 1) 
      results$data <- as.vector(results$data)
  }
  ##
  ## transforming data (Box - Cox)
  ##
  if (lambda != 1){
    if (lambda != 0)
      results$data <- (results$data * lambda + 1)^(1/lambda)
    else
      results$data <- exp(results$data)
    messa$transformation <- paste("grf: Data transformed (Box-Cox), for lambda =", lambda)
    if (messages.screen) 
      cat(messa$transformation); cat("\n")
  }
  ##
  ## preparing output
  ##
  if (messages.screen) 
    cat(paste("grf: End of simulation procedure. Number of realizations:", nsim, "\n"))
  results  <- c(results, list(cov.model = cov.model, 
                              nugget = nugget, cov.pars = cov.pars, kappa = kappa, 
                              lambda = lambda, messages = messa, method = method,
                              .Random.seed = rseed, call = call))
  class(results) <- c("grf", "geodata")
  return(results)
}

"grf.aux1" <-
  function (nst, nugget, sigmasq, phi, kappa, cov.model = c("exponential", 
                                                "matern", "gaussian", "spherical", "circular", "cubic", "wave", "powered.exponential", 
                                                "cauchy", "gneiting", "gneiting.matern", "pure.nugget")) 
{
  cov.model <- match.arg(cov.model)
  cov.nst <- paste("grf: process with ", nst, " covariance structure(s)")
  cov.nugget <- paste("grf: nugget effect is: tausq=", nugget, sep="")
  cov.message <- NULL
  for (i in 1:nst) {
    if (phi[i] == 0) 
      cov.message[i] <- paste("grf: covariance model", i, "is a pure nugget effect")
    else {
      if (cov.model == "matern" | cov.model == "powered.exponential" | 
          cov.model == "cauchy" | cov.model == "gneiting-matern") 
        cov.message[i] <- paste("grf: covariance model ", 
                                i, " is: ", cov.model, "(sigmasq=", sigmasq[i], 
                                ", phi=", phi[i], ", kappa = ", kappa, ")", sep = "")
      else cov.message[i] <- paste("grf: covariance model ", 
                                   i, " is: ", cov.model, "(sigmasq=", sigmasq[i], 
                                   ", phi=", phi[i], ")", sep = "")
    }
  }
  return(list(nst = cov.nst, nugget = cov.nugget, cov.structures = cov.message))
}
"grf.aux2" <-
  function (xlim, ylim, step, cov.model, phi, kappa = NULL) 
{
  if(!is.null(kappa))
    if (cov.model == "matern" & kappa == 0.5) {
      cov.model <- "exponential"
    }
  cs <- switch(cov.model,
               spherical = c(1,1),
               gneiting.matern = c(2, kappa),
               powered.exponential = c(3,kappa),
               exponential = c(3,1),
               gaussian = c(3,2),
               matern = c(4, kappa),
               gneiting = c(5,1),
               cauchy = c(6,kappa),
               wave = c(7,1))
  covfct <- cs[1]
  shape <- cs[2]
  parameter <- c(phi, shape, 1)
  nx <- c(diff(xlim)/step + 1, diff(ylim)/step + 1)
  storage.mode(nx) <- "integer"
  ln <- 2
  storage.mode(ln) <- "integer"
  covnr <- covfct
  storage.mode(covnr) <- "integer"
  simustep <- step
  storage.mode(simustep) <- "double"
  param <- parameter
  storage.mode(param) <- "double"
  lparam <- length(parameter)
  storage.mode(lparam) <- "integer"
  res <- double(nx[1] * nx[2])
  mm <- integer(ln)
  x <- .C("woodandchan", covnr, nx, ln, simustep, param, lparam, 
          res = res, m = mm)$res
  cat("\n")
  return(x)
}
"grf.aux3" <-
  function (nsim, temp) 
{
  realiz <- rep(0, temp$n)
  for (i in 1:temp$nst) {
    realiz <- realiz + sqrt(temp$sigmasq[i]) * grf.aux2(xlim = temp$xlims, 
                                                        ylim = temp$ylims, step = temp$stepx, cov.model = temp$cov.model, 
                                                        phi = temp$phi[i], kappa = temp$kappa)
    NULL
  }
  return(realiz)
}
"grfclass" <-
  function (xy, sim, theta, kappa = NULL, estpars = FALSE) 
{
  xyzsim <- cbind(xy, sim)
  nsim <- dim(xyzsim)[2] - 2
  if (estpars == TRUE) {
    simpars <- matrix(0, nsim, 4)
    for (i in 1:nsim) {
      k <- i + 2
      pars <- trends.lik(xy, xyzsim[, k], ini = theta, 
                         kappa = kappa, residuals = FALSE)
      pars <- c(pars$theta, kappa = kappa)
      simpars[i, ] <- pars
    }
    results <- list(xyzsim = xyzsim, simpars = simpars, theta = theta, 
                    kappa = kappa)
  }
  else results <- list(xyzsim = xyzsim, theta = theta, kappa = kappa)
  class(results) <- "grf"
  return(results)
}

"image.grf" <-
  function (obj, sim.number = 1, ...) 
{
  x <- as.numeric(levels(as.factor(obj$coords[, 1])))
  nx <- length(x)
  y <- as.numeric(levels(as.factor(obj$coords[, 2])))
  ny <- length(y)
  obj$data <- as.matrix(obj$data)
  n <- nrow(obj$data)
  if (nx * ny != n) 
    stop("Probably irregular grid")
  m <- matrix(obj$data[, sim.number], ncol = ny)
  image(x, y, m, ...)
  return(invisible())
}

"ksline" <-
  function (geodata, coords=geodata$coords, data=geodata$data,
            locations, cov.pars = stop("covariance parameters (sigmasq and phi) needed"), 
            nugget = 0, micro.scale = 0,
            cov.model = c("exponential", "matern", "gaussian", 
              "spherical", "circular", "cubic", "wave",
              "powered.exponential", "cauchy", 
              "gneiting", "gneiting.matern", "pure.nugget"), kappa = 0.5,
            lambda = 1, n.samples.backtransform = 500, m0 = "ok", 
            nwin = "full", trend = 1, d = 2, ktedata = NULL, ktelocations = NULL,
            aniso.pars = FALSE, 
            signal = FALSE,  messages.screen = TRUE, dist.epsilon = 1e-10) 
{
  require(mva)
  call <- match.call()
  if(lambda != 1) {
    if(messages.screen)
      cat("ksline: Data transformation (Box-Cox) performed.\n")
    if(lambda == 0)
      data <- log(data)
    else data <- ((data^lambda) - 1)/lambda
  }
  cov.model <- match.arg(cov.model)
  coords <- as.matrix(coords)
  locations <- as.matrix(locations)
  dimnames(coords) <- list(NULL, NULL)
  dimnames(locations) <- list(NULL, NULL)
  if (!is.null(ktedata) & !is.null(ktelocations) & m0 != "kte"){
    cat("ksline: external variable (covariate) provided. Kriging ste to KTE\n")
    m0 <- "kte"
  }
  ##
  ## Correcting anisotropy
  ##
  if(any(aniso.pars) == TRUE) {
    coords.c <- coords.aniso(coord = coord, aniso.pars = aniso.pars
                             )
    locations.c <- coords.aniso(coord = locations, aniso.pars = 
                                aniso.pars)
  }
  else {
    coords.c <- coords
    locations.c <- locations
  }
  ## 2. Preparing KTE matrices #####
  ##  
  if (m0 == "kte") {
    ktedata <- as.matrix(ktedata)
    ktelocations <- as.matrix(ktelocations)
    dimnames(ktedata) <- list(NULL, NULL)
    dimnames(ktelocations) <- list(NULL, NULL)
  }
  n <- length(data)
  ni <- length(locations[, 1])
  tausq <- nugget
  sigmasq <- cov.pars[1]
  phi <- cov.pars[2]
  if (nwin == "full") {
    est <- rep(0, ni)
    dif <- rep(0, ni)
    kvar <- rep(0, ni)
    sumw <- rep(0, ni)
    wofmean <- rep(0, ni)
    iv <- varcov.spatial(coords = coords, cov.model = cov.model, 
                         kappa = kappa, nugget = nugget, cov.pars = cov.pars, 
                         inv = TRUE, det = FALSE, func.inv = "cholesky", scaled = FALSE)
    av <- mean(data)
    sd <- sqrt(var(data))
    one <- rep(1, n)
    tone <- t(one)
    toneiv <- crossprod(one, iv)
    den <- solve(toneiv %*% one)
    ml <- den %*% toneiv %*% data
    kmsd <- sqrt(den)
    means <- c(average = av, stdev = sd, kmean = ml, kmsd = kmsd)
    if (m0 != "kt") {
      mktlocations <- "Constant trend"
      beta <- ml
    }
    else {
      mktlocations <- rep(0, ni)
      if (m0 == "kt" & trend == 1) {
        if (d == 1) {
          xmat <- cbind(rep(1, n), coords[, 2])
          xmati <- cbind(rep(1, ni), locations[, 2])
        }
        else {
          xmat <- cbind(rep(1, n), coords[, 1], coords[, 2])
          xmati <- cbind(rep(1, ni), locations[, 1], locations[, 
                                                               2])
        }
        iviv <- solve(crossprod(xmat,iv) %*% xmat)
        txiv <- t(xmat) %*% iv
        beta <- iviv %*% txiv %*% data
        mkt <- xmat %*% beta
      }
      if (m0 == "kt" & trend == 2) {
        if (d == 1) {
          xmat <- cbind(rep(1, n), coords[, 2], (coords[, 2])^2)
          xmati <- cbind(rep(1, ni), locations[, 2], (locations[, 
                                                                2])^2)
        }
        else {
          xmat <- cbind(rep(1, n), coords[, 1], coords[, 2], 
                        (coords[, 1])^2, (coords[, 2])^2, coords[, 1] * coords[, 
                                                                               2])
          xmati <- cbind(rep(1, ni), locations[, 1], locations[, 
                                                               2], (locations[, 1])^2, (locations[, 2])^2, locations[, 
                                                                                                                     1] * locations[, 2])
        }
        iviv <- solve(crossprod(xmat,iv) %*% xmat)
        txiv <- crossprod(xmat,iv)
        beta <- iviv %*% txiv %*% data
        mkt <- xmat %*% beta
      }
    }
    if (m0 != "kte") 
      mktelocations <- "No external trend"
    else {
      if (m0 == "kte") {
        mktelocations <- rep(0, ni)
        xmat <- cbind(rep(1, n), ktedata)
        xmati <- cbind(rep(1, ni), ktelocations)
        iviv <- solve(crossprod(xmat,iv) %*% xmat)
        txiv <- crossprod(xmat,iv)
        beta <- iviv %*% txiv %*% data
        mkte <- xmat %*% beta
      }
    }
    for (i in 1:ni) {
      if (messages.screen) {
        if (ni < 11) 
          cat(paste("ksline: kriging location: ", i, "out of", 
                    ni, "\n"))
        else {
          if (ni < 101 & (i%%10 == 1)) 
            cat(paste("ksline: kriging location: ", i, "out of", 
                      ni, "\n"))
          if (ni > 100 & i%%100 == 1) 
            cat(paste("ksline: kriging location: ", i, "out of", 
                      ni, "\n"))
          if (i == ni) 
            cat(paste("ksline: kriging location: ", i, "out of", 
                      ni, "\n"))
        }
      }
      coords0 <- cbind((coords[, 1] - locations[i, 1]), (coords[, 2] - locations[i, 
                                                                                 2]))
      dm0 <- sqrt(coords0[, 1]^2 + coords0[, 2]^2)
      v0 <- cov.spatial(obj = dm0, cov.model = cov.model, 
                        kappa = kappa, cov.pars = cov.pars)
      v0[dm0 < dist.epsilon] <- micro.scale + sigmasq
      tv0 <- t(v0)
      v0iv <- crossprod(v0, iv)
      v0ivv0 <- v0iv %*% v0
      skw <- crossprod(v0,iv)
      wofmean[i] <- 1 - sum(skw)
      ##
      ## 4.2.1 Simple kriging with known mean
      ##
      if (is.numeric(m0) == TRUE) {
        dif[i] <- skw %*% (data - m0)
        est[i] <- m0 + dif[i]
        if (signal == TRUE) 
          kvar[i] <- sigmasq - v0ivv0
        else kvar[i] <- tausq + sigmasq - v0ivv0
        sumw[i] <- sum(skw)
      }
      ##
      ## 4.2.2 Simple kriging with data average mean
      ##
      if (m0 == "av") {
        dif[i] <- skw %*% (data - av)
        est[i] <- av + dif[i]
        if (signal == TRUE) 
          kvar[i] <- sigmasq - v0ivv0
        else kvar[i] <- tausq + sigmasq - v0ivv0
        sumw[i] <- sum(((tone/n) + skw - ((skw %*% one %*% 
                                           tone)/n)))
      }
      ##
      ## 4.2.3 Ordinary kriging (or SK with G.L.S. mean)
      ##
      if(m0 == "ok") {
        dif[i] <- skw %*% (data - ml)
        est[i] <- ml + dif[i]
        redu <- as.vector(1 - toneiv %*% v0)
        if(signal == TRUE)
          kvar[i] <- sigmasq - v0ivv0 + (redu %*%
                                         den %*% redu)
        else kvar[i] <- tausq + sigmasq - v0ivv0 + (
                                                    redu %*% den %*% redu)
        sumw[i] <- sum((den %*% one + tv0 - v0iv %*% 
                        one %*% den %*% tone) %*% iv)
      }
      ##
      ## 4.2.4 Universal Kriging (or Kriging with trend model) 
      ##
      if(m0 == "kt") {
        dif[i] <- skw %*% (data - mkt)
        est[i] <- xmati[i,  ] %*% beta + dif[i]
        redu <- as.vector(xmati[i,  ]) - as.vector(
                                                   txiv %*% v0)
        if(signal == TRUE)
          kvar[i] <- sigmasq - v0ivv0 + (redu %*%
                                         iviv %*% redu)
        else kvar[i] <- tausq + sigmasq - v0ivv0 + (
                                                    redu %*% iviv %*% redu)
        sumw[i] <- sum(skw + xmati[i,  ] %*% iviv %*%
                       txiv - skw %*% xmat %*% iviv %*% txiv)
        mktlocations[i] <- xmati[i,  ] %*% beta
      }
      ##
      ## 4.2.5 Kriging with external trend 
      ##
      if(m0 == "kte") {
        dif[i] <- skw %*% (data - mkte)
        est[i] <- xmati[i,  ] %*% beta + dif[i]
        redu <- as.vector(xmati[i,  ]) - as.vector(
                                                   txiv %*% v0)
        if(signal == TRUE)
          kvar[i] <- sigmasq - v0ivv0 - (redu %*%
                                         iviv %*% redu)
        else kvar[i] <- tausq + sigmasq - v0ivv0 + (
                                                    redu %*% iviv %*% redu)
        sumw[i] <- sum(skw + xmati[i,  ] %*% iviv %*%
                       txiv - skw %*% xmat %*% iviv %*% txiv)
        mktelocations[i] <- xmati[i,  ] %*% beta
      }
    }
    message <- "Kriging performed in global neighbourhood"
    if (messages.screen) 
      cat(paste(message,"\n"))
    results <- list(predict = est, krige.var = kvar, dif = dif, summary = means, 
                    ktrend = mktlocations, ktetrend = mktelocations, beta = beta, 
                    wofmean = wofmean)
  }
  else {
    nwin <- min(n, nwin)
    avwin <- rep(0, ni)
    sdwin <- rep(0, ni)
    mlwin <- rep(0, ni)
    kmsdwin <- rep(0, ni)
    estwin <- rep(0, ni)
    difwin <- rep(0, ni)
    kvarwin <- rep(0, ni)
    sumwwin <- rep(0, ni)
    wofmean <- rep(0, ni)
    if (m0 != "kt") 
      mkt <- "Constant position trend"
    else mkt <- rep(0, ni)
    if (m0 != "kte") 
      mkte <- "No external trend"
    else mkte <- rep(0, ni)
    if (m0 != "kt" & m0 != "kte") 
      betawin <- "No polynomial or external trend"
    if (m0 == "kt") {
      if (trend == 1) {
        if (d == 1) 
          xmati <- cbind(rep(1, ni), locations[, 2])
        else xmati <- cbind(rep(1, ni), locations[, 1], locations[, 
                                                                  2])
      }
      if (trend == 2) {
        if (d == 1) 
          xmati <- cbind(rep(1, ni), locations[, 2], locations[, 
                                                               2]^2)
        else xmati <- cbind(rep(1, ni), locations[, 1], locations[, 
                                                                  2], (locations[, 1])^2, (locations[, 2])^2, locations[, 1] * 
                            locations[, 2])
      }
      betawin <- matrix(0, nrow = (ncol(xmati) * ni), ncol = ncol(xmati))
    }
    if (m0 == "kte") {
      xmati <- cbind(rep(1, ni), ktelocations)
      if (is.vector(ktedata) == TRUE) 
        betawin <- matrix(0, nrow = (2 * ni), ncol = 2)
      else betawin <- matrix(0, nrow = ((ncol(ktedata) + 
                                         1) * ni), ncol = (ncol(ktedata) + 1))
    }
    for (i in 1:ni) {
      temp.win <- ksline.aux.1(coords = coords, coords.c = 
                               coords.c, data = data, n = n, locations = 
                               locations[i,  ], locations.c = locations.c[
                                                  i,  ],
                               cov.pars = cov.pars, nugget = nugget,
                               cov.model = cov.model, kappa = kappa, m0 = m0,
                               nwin = nwin, trend = trend, d = d, ktedata = 
                               ktedata, ktelocations = ktelocations, micro.scale
                               = micro.scale, 
                               location.number = i, xmati = xmati[i,  ], mkte
                               = NULL, mkt = NULL, betawin = NULL, signal = 
                               signal, dist.epsilon = dist.epsilon)
      avwin[i] <- temp.win$avwin
      sdwin[i] <- temp.win$sdwin
      mlwin[i] <- temp.win$mlwin
      kmsdwin[i] <- temp.win$kmsdwin
      estwin[i] <- temp.win$estwin
      difwin[i] <- temp.win$difwin
      kvarwin[i] <- temp.win$kvarwin
      sumwwin[i] <- temp.win$sumwwin
      wofmean[i] <- temp.win$wofmean
      if (m0 == "kt") 
        mkt[i] <- temp.win$mkt
      if (m0 == "kte") 
        mkte[i] <- temp.win$mkte
      if (m0 == "kt" | m0 == "kte") 
        betawin[i, ] <- temp.win$betawin
      if (messages.screen) {
        if (ni < 11) 
          cat(paste("ksline: kriging location: ", i, "out of", 
                    ni, "\n"))
        else {
          if (ni < 101 & (i%%10 == 1)) 
            cat(paste("ksline: kriging location: ", i, "out of", 
                      ni, "\n"))
          if (ni > 100 & i%%100 == 1) 
            cat(paste("ksline: kriging location: ", i, "out of", 
                      ni, "\n"))
          if (i == ni) 
            cat(paste("ksline: kriging location: ", i, "out of", 
                      ni, "\n"))
        }
      }
    }
    message <- "kriging performed in moving neighbourhood"
    if (messages.screen) 
      cat(paste(message,"\n"))
    results <- list(predict = estwin, krige.var = kvarwin, dif = difwin, 
                    avtrend = avwin, sd = sdwin, oktrend = mlwin, oksd = kmsdwin, 
                    ktrend = mkt, ktetrend = mkte, beta = betawin, sumw = sumwwin, 
                    wofmean = wofmean)
  }  
  if(lambda != 1) {
    if(messages.screen)
      cat("Back-transforming the predictions according to the (Box-Cox) parameter lambda\n")
    if(lambda == 0) {
      predict.transf <- results$predict
      results$predict <- exp(predict.transf) - 0.5 * results$krige.var
      results$krige.var <- (exp(2 * predict.transf - results$krige.var)) * (exp(results$krige.var) - 1)
    }
    if(lambda > 0) {
      if(messages.screen)
        cat("Back-transformation done by sampling from the resulting (normal) predictive distribution\n")
      ap.warn <- options()$warn
      options(warn = -1)
      temp.data <- matrix(rnorm(ni * n.samples.backtransform,
                                mean = results$predict, sd = sqrt(results$krige.var)),
                          nrow = ni)
      options(warn = ap.warn)
      temp.data[(results$krige.var == 0),  ] <- results$predict[(results$krige.var == 0)]
      temp.data[temp.data < -1/lambda] <- -1/lambda     
      temp.data <- ((temp.data * lambda) + 1)^(1/lambda)
#      temp.data[is.na(temp.data)] <- Inf
      results$predict <- as.vector(apply(temp.data, 1, mean))
      results$krige.var <- as.vector(apply(temp.data, 1, var))
    }
    if(lambda < 0) {
      cat("Resulting distribution has no mean for lambda < 0 - back transformation not performed\n"
          )
    }
  }
  results$locations <- locations
  results$message <- message
  results$call <- call
  class(results) <- c("kriging")
  return(invisible(results))
}

"ksline.aux.1" <-
  function (coords, data, locations, n, cov.pars, nugget, cov.model, kappa, 
            m0, nwin, trend, d, ktedata, ktelocations, mbased, location.number, 
            xmati, mkte, mkt, betawin, signal) 
{
  require(mva)
  i <- location.number
  sigmasq <- cov.pars[1]
  phi <- cov.pars[2]
  tausq <- nugget
  coords0 <- cbind((coords.c[, 1] - locations.c[1]), (coords.c[, 2] -
                                                      locations.c[2]))
  dm0 <- sqrt(coords0[, 1]^2 + coords0[, 2]^2)
  coordswin <- coords[order(dm0)[1:nwin],  ]
  coordswin.c <- coords.c[order(dm0)[1:nwin],  ]
  datawin <- data[order(dm0)[1:nwin]]
  ivwin <- varcov.spatial(coords = coordswin.c, cov.model = cov.model,
                          kappa = kappa, nugget = nugget, cov.pars = cov.pars, inv = TRUE,
                          det = FALSE, func.inv = "cholesky", scaled = FALSE, only.decomp = FALSE)
  avwin <- mean(datawin)
  sdwin <- sqrt(var(datawin))
  onewin <- rep(1, nwin)
  toneivwin <- crossprod(onewin, ivwin)
  denwin <- solve(toneivwin %*% onewin)
  mlwin <- denwin %*% toneivwin %*% datawin
  kmsdwin <- sqrt(denwin)
  coords0win <- cbind((coordswin[, 1] - locations[1]), (coordswin[, 2] -
                                                        locations[2]))
  coords0win.c <- cbind((coordswin.c[, 1] - locations.c[1]), (coordswin.c[
                                                                          , 2] - locations.c[2]))
  dm0win <- sqrt(coords0win.c[, 1]^2 + coords0win.c[, 2]^2)
  v0win <- cov.spatial(obj = dm0win, cov.model = cov.model, kappa = kappa,
                       cov.pars = cov.pars)
  v0win[dm0win < dist.epsilon] <- micro.scale + sigmasq
  skwwin <- crossprod(v0win, ivwin)
  wofmean <- 1 - sum(skwwin)
  if(m0 == "kt" & trend == 1) {
    if(d == 1)
      xmatwin <- cbind(rep(1, nwin), coordswin[, 2])
    else xmatwin <- cbind(rep(1, nwin), coordswin[, 1], coordswin[
                                                                  , 2])
    txivwin <- crossprod(xmatwin, ivwin)
    ivivwin <- solve(txivwin %*% xmatwin)
    betawin <- ivivwin %*% txivwin %*% datawin
    mktwin <- xmatwin %*% betawin
    mkt <- xmati %*% betawin
  }
  if(m0 == "kt" & trend == 2) {
    if(d == 1)
      xmatwin <- cbind(rep(1, nwin), coordswin[, 2], (
                                                      coordswin[, 2])^2)
    else xmatwin <- cbind(rep(1, nwin), coordswin[, 1], (coordswin[
                                                                   , 1])^2, coordswin[, 2], (coordswin[, 2])^
                          2, coordswin[, 1] * coordswin[, 2])
    xmatwin.cent <- xmatwin
    xmatwin.cent[, 2] <- xmatwin.cent[, 2] - mean(xmatwin[, 2])
    xmatwin.cent[, 3] <- xmatwin.cent[, 3] - mean(xmatwin[, 3])
    ivivwin <- solve(crossprod(xmatwin.cent, ivwin) %*% 
                     xmatwin.cent)
    txivwin <- crossprod(xmatwin.cent, ivwin)
    betawin <- ivivwin %*% txivwin %*% datawin
    betawin <- mean(datawin) - crossprod(betawin, c(0, mean(xmatwin[
                                                                    , 2]), mean(xmatwin[, 3])))
    mktwin <- xmatwin %*% betawin
    mkt <- xmati %*% betawin
  }
  if(m0 == "kte") {
    if(is.vector(ktedata))
      ktedatawin <- ktedata[order(dm0)[1:nwin]]
    else ktedatawin <- ktedata[order(dm0)[1:nwin],  ]
    xmatwin <- cbind(rep(1, nwin), ktedatawin)
    ivivwin <- solve(crossprod(xmatwin, ivwin) %*% xmatwin)
    txivwin <- crossprod(xmatwin, ivwin)
    betawin <- ivivwin %*% txivwin %*% datawin
    mktewin <- xmatwin %*% betawin
    mkte <- xmati %*% betawin
  }
  ##
  ##  Simple kriging with known mean
  ##
  if(is.numeric(m0)) {
    difwin <- skwwin %*% (data - m0)
    estwin <- m0win + difwin
    if(signal)
      kvarwin <- sigmasq - crossprod(v0win, ivwin) %*% v0win
    else kvarwin <- tausq + sigmasq - crossprod(v0win, ivwin) %*%
      v0win
    sumwwin <- sum(skwwin)
  }
  ##
  ## 4.2.2 Simple kriging with data average mean
  ##
  if(m0 == "av") {
    difwin <- skwwin %*% (datawin - avwin)
    estwin <- avwin + difwin
    if(signal)
      kvarwin <- sigmasq - crossprod(v0win, ivwin) %*% v0win
    else kvarwin <- tausq + sigmasq - crossprod(v0win, ivwin) %*%
      v0win
    sumwwin <- sum(((t(onewin)/nwin) + skwwin - ((skwwin %*% onewin %*%
                                                  t(onewin))/n)))
  }
  ##
  ## Ordinary kriging (or SK with G.L.S. mean)
  ##
  if(m0 == "ok") {
    difwin <- skwwin %*% (datawin - mlwin)
    estwin <- mlwin + difwin
    redu <- as.vector(1 - toneivwin %*% v0win)
    if(signal)
      kvarwin <- sigmasq - v0win %*% ivwin %*% v0win + (
                                                        redu %*% denwin %*% redu)
    else kvarwin <- tausq + sigmasq - v0win %*% ivwin %*% v0win +
      (redu %*% denwin %*% redu)
    sumwwin <- sum((denwin %*% onewin + t(v0win) - crossprod(v0win,
                                                             ivwin) %*% onewin %*% denwin %*% t(onewin)) %*% ivwin)
  }
  ##
  ## Universal Kriging (or Kriging with trend model) 
  ##
  if(m0 == "kt") {
    difwin <- skwwin %*% (datawin - mktwin)
    estwin <- mkt + difwin
    xmati <- as.vector(xmati)
    redu <- as.vector(xmati) - as.vector(txivwin %*% v0win)
    if(signal)
      kvarwin <- sigmasq - v0win %*% ivwin %*% v0win + (
                                                        redu %*% ivivwin %*% redu)
    else kvarwin <- tausq + sigmasq - v0win %*% ivwin %*% v0win +
      (redu %*% ivivwin %*% redu)
    sumwwin <- sum(skwwin + xmati %*% ivivwin %*% txivwin - skwwin %*%
                   xmatwin %*% ivivwin %*% txivwin)
  }
  ##
  ## Kriging with external trend 
  ##
  if(m0 == "kte") {
    difwin <- skwwin %*% (datawin - mktewin)
    estwin <- mkte + difwin
    xmati <- as.vector(xmati)
    redu <- as.vector(xmati) - as.vector(txivwin %*% v0win)
    if(signal)
      kvarwin <- sigmasq - v0win %*% ivwin %*% v0win + (
                                                        redu %*% ivivwin %*% redu)
    else kvarwin <- tausq + sigmasq - v0win %*% ivwin %*% v0win +
      (redu %*% ivivwin %*% redu)
    sumwwin <- sum(skwwin + xmati %*% ivivwin %*% txivwin - skwwin %*%
                   xmatwin %*% ivivwin %*% txivwin)
  }
  ##
  ##
  ##  
  results <- list(avwin = avwin, sdwin = sdwin, mlwin = mlwin, kmsdwin = 
                  kmsdwin, wofmean = wofmean, betawin = betawin, mkt = mkt, mkte
                  = mkte, difwin = difwin, estwin = estwin, kvarwin = kvarwin,
                  sumwwin = sumwwin)
  return(results)
}

"likfit.nospatial" <-
  function(temp.list, ...)
{
  results <- list()
  z <- temp.list$z
  n <- temp.list$n
  npar.mean <- temp.list$npar.mean
  xmat <- temp.list$xmat
  txmat <- temp.list$txmat
  ixx <- solve(crossprod(xmat))
  if(temp.list$fix.lambda == FALSE){
    if (temp.list$minimisation.function == "nlm"){
      assign(".temp.lower.lambda",-2, pos=1)
      assign(".temp.upper.lambda", 2, pos=1)
      results <- nlm(proflik.lambda, 1, ...)
      if(exists(".temp.lambda")){
        results$lambda <- .temp.lambda
        remove(".temp.lambda", pos=1, inherits = TRUE)
      }
      else{
        results$lambda <- results$estimate
      }
      rm(.temp.lower.lambda, .temp.upper.lambda, inherits = TRUE, pos=1)
    }
    if (temp.list$minimisation.function == "nlmP"){
      results <- nlmP(proflik.lambda, 1, lower=-2, upper=2,...)  
      results$lambda <- results$estimate
    }
    if (temp.list$minimisation.function == "optim"){
      results <- optim(1, proflik.lambda, method="L-BFGS-B", lower=-2, upper=2,...)
      results$minimum <- results$value
      results$lambda <- results$par
    }
    if(results$lambda == 1) {
      temp.list$log.jacobian <- 0
    }
    else {
      if(any(z <= 0))
        stop("Transformation option not allowed when there are zeros or negative data")
      if(any(z^(results$lambda - 1) <= 0))
        temp.list$log.jacobian <- log(prod(z^(results$lambda - 1)))
      else temp.list$log.jacobian <- sum(log(z^(results$lambda - 1)))
      if(results$lambda == 0)
        z <- log(z)
      else z <- ((z^results$lambda) - 1)/results$lambda
    }
  }
  else{
    results$lambda <- temp.list$lambda
    results$code <- 1
    if (temp.list$minimisation.function == "optim") results$convergence <- 0
  }
  ssres <- (z %*% (diag(n) - xmat %*%
                   solve(crossprod(xmat)) %*% txmat) %*% z)
  if(temp.list$method == "ML"){
    results$tausqhat <- ssres/n
    if(temp.list$fix.lambda)
      results$minimum <- as.vector(((n/2) * log(2 * pi) +
                          (n/2) * log(results$tausqhat) +
                          (n/2)  -
                          temp.list$log.jacobian))
  }
  if(temp.list$method == "RML") {
    results$tausqhat  <- (ssres/(n-npar.mean))
    if(temp.list$fix.lambda){
      results$minimum <- as.vector((((n - npar.mean)/2) * log(2 * pi) +
                          ((n - npar.mean)/2) * log(results$tausqhat) +
                          (n/2) -
                          temp.list$log.jacobian
                          ))
    }
  }
  if (temp.list$minimisation.function == "optim") results$value <- results$minimum    
  return(results)
}
"lines.grf" <-
function (obj, max.dist = max(dist(obj$coords)), length = 100, 
    lwd = 2, ...) 
{
    require(mva)
    if (obj$cov.model == "matern" | obj$cov.model == "powered.exponential" | 
        obj$cov.model == "cauchy" | obj$cov.model == "gneiting-matern") 
        kappa <- obj$kappa
    else kappa <- NULL
    distance <- seq(0, max.dist, length = length)
    if (is.vector(obj$cov.pars)) 
        sill.tot <- obj$nugget + obj$cov.pars[1]
    else sill.tot <- obj$nugget + sum(obj$cov.pars[, 1])
    gamma <- sill.tot - cov.spatial(distance, cov.model = obj$cov.model, 
        kappa = kappa, cov.pars = obj$cov.pars)
    lines(distance, gamma, lwd = lwd, ...)
}
"lines.variogram" <-
function (obj, max.dist = max(obj$u), type = "o", scaled = FALSE, ...) 
{
  if (scaled) 
    obj$v <- obj$v/obj$var.mark
  if (!is.matrix(obj$v)) 
    lines(obj$u[obj$u <= max.dist], obj$v[obj$u <= max.dist], 
          type = type, ...)
  else {
    for (j in 1:ncol(obj$v)) lines(obj$u[obj$u <= max.dist], 
                                   obj$v[obj$u <= max.dist, j], type = type, ...)
  }
}
"lines.variomodel" <-
function (obj, max.dist = obj$max.dist, n.points = 100, scaled = FALSE,...)
{
  if (scaled) 
    obj$v <- obj$v/obj$var.mark
  if (is.null(max.dist)) 
    stop("argument max.dist needed for this object")
  if (obj$cov.model == "matern" | obj$cov.model == "powered.exponential" | 
      obj$cov.model == "cauchy" | obj$cov.model == "gneiting-matern") 
    kappa <- obj$kappa
  else kappa <- NULL
  distance <- seq(0, max.dist, length = n.points)
  if (is.vector(obj$cov.pars)) 
    sill.tot <- obj$nugget + obj$cov.pars[1]
  else sill.tot <- obj$nugget + sum(obj$cov.pars[, 1])
  if (obj$cov.pars[2] < 1e-12)
    gamma <- rep(sill.tot, n.points)
  else
    gamma <- sill.tot - cov.spatial(distance, cov.model = obj$cov.model, 
                                    kappa = kappa, cov.pars = obj$cov.pars)
  lines(distance, gamma, ...)
}
"loglik.spatial" <-
function(pars)
{
  tausq <- pars[1]
  sigmasq <- pars[2]
  sill.tot <- tausq + sigmasq
  phi <- pars[3]
  lambda <- pars[4]
  z <- .temp.list$z
  n <- .temp.list$n
  if(.temp.list$fix.lambda == FALSE) {
    if(lambda == 1) {
      .temp.list$log.jacobian <- 0
    }
    else {
      if(any(z < 0))
        stop("Transformation option not allowed when there are zeros or negative data"
             )
      if(any(z^(lambda - 1) <= 0))
        .temp.list$log.jacobian <- log(prod(z^(lambda -
                                              1)))
      else .temp.list$log.jacobian <- sum(log(z^(lambda - 1)))
      if(lambda == 0)
        z <- log(z)
      else z <- ((z^lambda) - 1)/lambda
    }
  }
  npar.mean <- .temp.list$npar.mean
  kappa <- .temp.list$kappa
  xmat <- .temp.list$xmat
  txmat <- .temp.list$txmat
  covinf <- varcov.spatial(dists.lowertri = .temp.list$
                           dists.lowertri, cov.model = .temp.list$cov.model,
                           kappa = kappa, nugget = tausq, cov.pars = c(
                                                            sigmasq, phi), scale = FALSE, inv = TRUE, det = TRUE,
                           sqrt.inv = TRUE)
  six <- crossprod(covinf$sqrt.inverse, xmat)
  xix <- crossprod(six)
  if(length(as.vector(xix)) == 1) {
    ixix <- 1/xix
    choldet <- 0.5 * log(xix)
  }
  else {
    chol.xix <- chol(xix)
    ixix <- chol2inv(chol.xix)
    choldet <- sum(log(diag(chol.xix)))
  }
  xi <- crossprod(xmat, covinf$inverse)
  ssresmat <- z %*% (covinf$inverse - crossprod(xi, ixix) %*%
                     xi) %*% z
  if(.temp.list$method == "ML") {
    loglik <- ( - (n/2) * log(2 * pi) -
               covinf$log.det.to.half -
               0.5 * ssresmat + 
               .temp.list$log.jacobian)
  }
  if(.temp.list$method == "RML") {
    xx.eigen <- eigen(crossprod(xmat), symmetric = TRUE, only.values = TRUE)
    loglik <- ( - ((n - npar.mean)/2) * log(2 * pi) +
               0.5 * sum(log(xx.eigen$values)) -
               covinf$log.det.to.half -
               (0.5) * ssresmat -
               choldet +
               .temp.list$log.jacobian)
  }
  return(as.vector(loglik))
}
"loss.olsvario" <-
function (theta) 
{
  if (exists(".temp.theta", w=1)) remove(".temp.theta", pos=1, inherits = TRUE)
  theta.minimiser <- theta
  penalty <- 10000 * sum(.global.lower - pmin(theta, .global.lower))
  theta <- pmax(theta, .global.lower)
  if (any(theta.minimiser < .global.lower))
    assign(".temp.theta", theta, pos=1)
  u <- .global.list$data$u
  v <- .global.list$data$v
  kappa <- .global.list$kappa
  if (.global.list$fix.nugget == FALSE) {
    tausq <- theta[1]
    sigmasq <- theta[2]
    phi <- theta[3]
    tau1 <- tausq
  }
  else {
    tausq <- .global.list$nugget
    sigmasq <- theta[1]
    phi <- theta[2]
    tau1 <- tausq + 1
  }
  sill.tot <- sigmasq + tausq
  if (.global.list$fix.nugget == FALSE) {
    gamma <- sill.tot - cov.spatial(u, cov.model = .global.list$cov.model, 
                                    kappa = kappa, cov.pars = c(theta[2], theta[3]))
  }
  else {
    gamma <- sill.tot - cov.spatial(u, cov.model = .global.list$cov.model, 
                                    kappa = kappa, cov.pars = c(theta[1], theta[2]))
  }
  loss <- sum((v - gamma)^2)
  return(loss + penalty)
}

"loss.wlsvario" <-
function (theta) 
{
  if (exists(".temp.theta", w=1)) remove(".temp.theta", pos=1, inherits = TRUE)
  theta.minimiser <- theta
  penalty <- 10000 * sum(.global.lower - pmin(theta, .global.lower))
  theta <- pmax(theta, .global.lower)
  if (any(theta.minimiser < .global.lower))
    assign(".temp.theta", theta, pos=1)
  u <- .global.list$data$u
  v <- .global.list$data$v
  n <- .global.list$data$n
  kappa <- .global.list$kappa
  if (.global.list$fix.nugget == FALSE) {
    tausq <- theta[1]
    sigmasq <- theta[2]
    phi <- theta[3]
    tau1 <- tausq
  }
  else {
    tausq <- .global.list$nugget
    sigmasq <- theta[1]
    phi <- theta[2]
    tau1 <- tausq + 1
  }
  sill.tot <- sigmasq + tausq
  if (.global.list$fix.nugget == FALSE) {
    gamma <- sill.tot - cov.spatial(u, cov.model = .global.list$cov.model, 
                                    kappa = kappa, cov.pars = c(theta[2], theta[3]))
  }
  else {
    gamma <- sill.tot - cov.spatial(u, cov.model = .global.list$cov.model, 
                                    kappa = kappa, cov.pars = c(theta[1], theta[2]))
  }
  if (.global.list$weight == "npairs") 
    loss <- sum(n * (v - gamma)^2)
  if (.global.list$weight == "cressie") 
    loss <- sum((n/(gamma^2)) * (v - gamma)^2)
  return(loss + penalty)
}
"matern" <-
  function (u, phi, kappa) 
{
  if(is.vector(u)) names(u) <- NULL
  if(is.matrix(u)) dimnames(u) <- list(NULL, NULL)
  uvec <- u/phi
  ucov <- ifelse(uvec > 0, ((((2^(kappa-1))*gamma(kappa))^(-1))*(uvec^kappa)*besselK(x=uvec, nu=kappa)), 1)    
  ## The following needs more checking:
  ##        min.non.inf <- min(ucov[ucov != Inf])
  ##        dist.min.non.inf <- min(u[ucov <= 1.01 * min.non.inf])
  ##        ucov[u <= dist.min.non.inf] <- 0
  ##        print(c(phi, dist.min.non.inf, min.non.inf))        
  ##
  ucov[u > 600*phi] <- 0 
  return(ucov)
}
"nice" <-
  function (x, decimal = 2, fixed = FALSE) 
{
  ergb <- x
  index <- (x != 0) & is.finite(x)
  if (fixed) 
    n <- 0
  else n <- floor(log(abs(x[index]))/log(10))
  ergb[index] <- trunc(x[index]/10^(n - decimal)) * 10^(n - 
                                                        decimal)
  return(ergb)
}
"nlmP" <-
  function(objfunc, params, lower = rep( -Inf, length(params)), upper = rep(
                                                                  +Inf, length(params)), ... )
{
  ## minimizer, using nlm with transformation of variables
  ##
  ## objfunc is a function to be optimised
  ## params is a starting value for the parameters
  Nparams <- length(params)
  if(length(lower) != Nparams)
    stop(" lower boundry different length than params")
  if(length(upper) != Nparams)
    stop(" upper boundry different length than params")
  checklimits <- upper - lower
  if(any(checklimits <= 0))
    stop(" bad boundries")
  if(any(params < lower))
    stop(" starting params too low")
  if(any(params > upper))
    stop(" starting params too high")
  
  bothlimQQ <- (lower != (-Inf)) & (upper != +Inf)
  loweronlyQQ <- (lower != (-Inf)) & (upper == +Inf)
  upperonlyQQ <- (lower == (-Inf)) & (upper != +Inf)
  ubothQQ <- upper[bothlimQQ]
  lbothQQ <- lower[bothlimQQ]
  dbothQQ <- ubothQQ - lbothQQ
  loneQQ <- lower[loweronlyQQ]
  uoneQQ <- upper[upperonlyQQ]
  
  .bounds.list <- list(bothlimQQ = bothlimQQ, 
                       loweronlyQQ = loweronlyQQ,
                       upperonlyQQ = upperonlyQQ,
                       ubothQQ = ubothQQ,
                       lbothQQ = lbothQQ,
                       dbothQQ = dbothQQ,
                       loneQQ = loneQQ,
                       uoneQQ = uoneQQ)
  
  assign(".objfuncQQ", objfunc, pos=1)
  assign(".bounds.list", .bounds.list, pos=1)
  
  ## reduce the parameter space by a scale to keep parameters
  ## away from the boundries
  
  normaltomad <- function(normalparamsX)
    {
      madparamsX <- normalparamsX
      if(any(.bounds.list$bothlimQQ)) {
        noughtone <- (normalparamsX[.bounds.list$bothlimQQ] -
                      .bounds.list$lbothQQ)/.bounds.list$dbothQQ
        madparamsX[.bounds.list$bothlimQQ] <- log(noughtone/(1 - noughtone))
      }
      
      if(any(.bounds.list$loweronlyQQ))
        madparamsX[.bounds.list$loweronlyQQ] <-
          log(normalparamsX[.bounds.list$loweronlyQQ] - .bounds.list$loneQQ)
      
      if(any(.bounds.list$upperonlyQQ))
        madparamsX[.bounds.list$upperonlyQQ] <-
          log(.bounds.list$uoneQQ - normalparamsX[.bounds.list$upperonlyQQ])
      
      return(madparamsX)
    }
  
  madtonormalQQ <<- function(madparamsX)
    {
      normalparamsX <- madparamsX
      
      if(any(.bounds.list$bothlimQQ)) {
###        madparamsX[((.bounds.list$bothlimQQ) & (madparamsX > 300))] <- 300
        emad <- exp(madparamsX[.bounds.list$bothlimQQ])
        normalparamsX[.bounds.list$bothlimQQ] <-
          .bounds.list$dbothQQ * (emad/(1 + emad)) + .bounds.list$lbothQQ
      }
      
      if(any(.bounds.list$loweronlyQQ)){
        normalparamsX[.bounds.list$loweronlyQQ] <-
          exp(madparamsX[.bounds.list$loweronlyQQ]) + .bounds.list$loneQQ
      }
      
      if(any(.bounds.list$upperonlyQQ))
        normalparamsX[.bounds.list$upperonlyQQ] <-
          - exp(madparamsX[.bounds.list$upperonlyQQ]) + .bounds.list$uoneQQ
      
      if(exists(".ind.prof.phi"))
        if(is.nan(normalparamsX[.ind.prof.phi]))
          normalparamsX[.ind.prof.phi] <- 0
      
      return(normalparamsX)
    }
  
  newobjfunc <- function(madparams) {
    normalparams <-  madtonormalQQ(madparams)
    
    .objfuncQQ(normalparams)
    
  }
  
  startmadparams <- normaltomad(params)
  result <- nlm(newobjfunc, startmadparams, ...)
  result$madestimate <- result$estimate
  result$estimate <- madtonormalQQ(result$madestimate)
  remove(".bounds.list", pos=1, inherits=T)
  remove(".objfuncQQ", pos=1, inherits=T)
  remove("madtonormalQQ", pos=1, inherits=T)
  
###  return(result, madtonormalQQ(normaltomad(params)),params)
  return(result)
}
"olsfit" <-
  function (vario, ini, cov.model = c("exponential", "matern", 
                          "gaussian", "spherical", "circular", "cubic", "wave", "powered.exponential", "cauchy", 
                          "gneiting", "gneiting.matern", "pure.nugget"), kappa = NULL, simul.number = NULL, 
            max.dist = "all", minimisation.function = "nlm", lower = 0, fix.nugget = FALSE, 
            nugget = 0, print.level=0, messages.screen=TRUE) 
{
  call <- match.call()
  if (minimisation.function != "nlm" & minimisation.function != "nls") 
    stop("minimization functions available are: nlm or nls")
  if (is.matrix(vario$v) & is.null(simul.number)) 
    stop("object in vario$v is a matrix. This function works for only 1 empirical variogram")
  if (!is.null(simul.number)) 
    vario$v <- vario$v[, simul.number]
  if(is.matrix(ini)) {
    inilength <- dim(ini)[2]
    if(fix.nugget == FALSE & inilength != 3)
      stop("wrong number of columns for ini (must be 3)")
    if(fix.nugget == TRUE & inilength != 2)
      stop("wrong number of columns for ini (must be 2)")
  }
  else {
    inilength <- length(ini)
    if (fix.nugget == FALSE & inilength != 3) 
      stop("wrong length for ini (must be 3)")
    if (fix.nugget == TRUE & inilength != 2) 
      stop("wrong length for ini (must be 2)")
  }
  cov.model <- match.arg(cov.model)
  if (max.dist == "all") 
    data <- data.frame(u = vario$u, v = vario$v)
  else data <- data.frame(u = vario$u[vario$u <= max.dist], v = vario$v[vario$u <= max.dist])
  data$kappa <- kappa
  data$cov.model <- cov.model
  if (minimisation.function == "nls") {
    require(nls)
    if (fix.nugget == FALSE) {
      result <- nls(v ~ ((sigmasq + tausq) - cov.spatial(u, 
                                                         cov.model = cov.model, kappa = kappa, cov.pars = c(sigmasq, 
                                                                                                 phi))), start = list(tausq = ini[1], sigmasq = ini[2], 
                                                                                                           phi = ini[3]), data = data)
      nugget <- result$parameter[1]
      cov.pars <- as.vector(result$parameter[2:3])
      value <- vario$v - ((nugget + cov.pars[1]) - cov.spatial(vario$u, 
                                                               cov.model = cov.model, kappa = kappa, cov.pars = cov.pars))
    }
    else {
      data$nugget <- nugget
      result <- nls(v ~ ((sigmasq + ftau) - cov.spatial(u, 
                                                        cov.model = cov.model, kappa = kappa, cov.pars = c(sigmasq, 
                                                                                                phi))), start = list(sigmasq = ini[1], phi = ini[2]), 
                    data = data)
      cov.pars <- as.vector(result$parameter[1:2])
      value <- vario$v - ((nugget + cov.pars[1]) - cov.spatial(vario$u, 
                                                               cov.model = cov.model, kappa = kappa, cov.pars = cov.pars))
    }
    value <- sum(value * value)
    message <- "nls does not provides convergence message"
  }
  if (minimisation.function == "nlm") {
    .global.list <<- list(data = data, fix.nugget = fix.nugget, nugget = nugget, 
                          kappa = kappa, cov.model = cov.model)
    .global.lower <<- lower
    if(is.matrix(ini) | is.data.frame(ini)) {
      ini <- as.matrix(ini)
      cat("olsfit: searching for the best initial value\n")
      ini.search <- ini
      dimnames(ini.search) <- list(NULL, NULL)
      ols.ini <- round(100000000. * apply(ini.search, 1, 
                                          loss.olsvario))
      ini <- as.vector(ini.search[ols.ini == min(ols.ini),
                                      ])
      cat("olsfit: best initial value:\n")
      if(fix.nugget == TRUE){
        names(ini) <- c("sill", "range")
        print(ini)
      }
      else{
        names(ini) <- c("nugget", "sill", "range")
        print(ini)
      }
    }
    result <- nlm(loss.olsvario, ini, iterlim =
                  10000,print.level=print.level)	   
    if (exists(".temp.theta", w=1)){
      result$estimate <- .temp.theta
      remove(".temp.theta", pos=1, inherits = TRUE)
    }
    rm(.global.list, .global.lower, inherits = TRUE)
    value <- result$minimum
    message <- paste("Exit code:", result$code)
    if (fix.nugget == FALSE) {
      nugget <- result$estimate[1]
      cov.pars <- as.vector(result$estimate[2:3])
    }
    else {
      cov.pars <- as.vector(result$estimate[1:2])
    }
  }
  dmax <- max(vario$u)
  estimation <- list(nugget = nugget, cov.pars = cov.pars, 
                     cov.model = cov.model, kappa = kappa, value = value, 
                     max.dist = dmax, minimisation.function = minimisation.function, message = message)
  estimation$method <- "OLS"
  estimation$call <- call
  class(estimation) <- "variomodel"
  if(messages.screen){
    cat("olsfit: estimated model parameters are:\n")
    cat(paste("covariance model:", cov.model))
    if(cov.model == "matern" | cov.model == "powered.exponential" | 
       cov.model == "cauchy" | cov.model == "gneiting.matern")
      cat(paste(" with kappa =", kappa))
    if(!is.null(kappa))
      if(cov.model == "matern" & kappa == 0.5)
        cat(" (exponential)")
    cat("\n")
    print(c(nugget=estimation$nugget, sill=estimation$cov.pars[1], range=estimation$cov.pars[2]))
  }
  return(estimation)
}
"plot.geodata" <-
  function (geodata, coords = geodata$coords, data = geodata$data, trend = "cte",
            lambda=1, col.data = 1, window.new = FALSE, ...) 
{
  par.ori <- par(no.readonly=TRUE)
  on.exit(par(par.ori))
  coords <- as.matrix(geodata$coords)
  data <- as.matrix(data)
  data <- data[, col.data]
  if (window.new) X11()
  if (lambda != 1){
    if(lambda == 0)
      data <- log(data)
    else
      data <- ((data^lambda)-1)/lambda
  }
  par(mfrow = c(2, 2))
  par(pty = "s")
  par(mar = c(4, 4, 0, .5))
  data.quantile <- quantile(data)
  plot(coords, xlab = "Coord X", ylab = "Coord Y", type = "n")
  if(length(unique(data.quantile)) == length(data.quantile)) {
    points(coords[(data <= data.quantile[2]), ], cex = 0.4, col = "blue")
    points(coords[((data > data.quantile[2]) & (data <= data.quantile[3])), 
                  ], cex = 0.6, col = "green")
    points(coords[((data > data.quantile[3]) & (data <= data.quantile[4])), 
                  ], cex = 0.9, col = "yellow")
    points(coords[(data > data.quantile[4]), ], cex = 1.2, col = "red")
  }
  else {
    points(coords, pch = 16)
  }
  par(pty = "m")
  par(mar = c(4, 4, 1, 1))
###  plot(variog(coords=coords, data=data, trend=trend,...))
  hist(data)
  plot(coords[, 1], data, xlab = "Coord X", cex = 1)
  plot(coords[, 2], data, xlab = "Coord Y", cex = 1)
  return(invisible())
}

"plot.grf" <-
  function (obj, model.line = TRUE, plot.grid = FALSE, ylim="default", ...) 
{
  nsim <- ncol(obj$data)
  if (plot.grid) 
    plot(obj$coords[, 1], obj$coords[, 2], xlab = "Coord X", ylab = "Coord Y")
  if (is.vector(obj$cov.pars)) 
    sill.tot <- obj$nugget + obj$cov.pars[1]
  else sill.tot <- obj$nugget + sum(obj$cov.pars[, 1])
  if (obj$lambda != 1){
    if (obj$lambda == 0) data <- log(obj$data)
    else data <- ((obj$data^obj$lambda)-1)/obj$lambda
  }
  else
    data <- obj$data          
  sim.bin <- variog(obj)
  plot(sim.bin, var.lines = FALSE, ylim = c(0,max(sim.bin$v, sill.tot)), ...)
  if (model.line){
    var.model <- list(nugget = obj$nugget, cov.pars = obj$cov.pars, 
                      kappa = obj$kappa, max.dist = max(sim.bin$u),
                      cov.model = obj$cov.model)
    lines.variomodel(var.model, lwd = 3, ...)
  }
  return(invisible())
}

"plot.variogram" <-
  function (obj, max.dist = max(obj$u), scaled = FALSE,
            var.lines = FALSE, envelope.obj = NULL,
            bin.cloud = FALSE, ylim = "default", ...) 
{
  if(!exists("type", envir=sys.frame(1))){
    if (obj$output.type == "bin") type <- "b"
    if (obj$output.type == "smooth") type <- "l"
    if (obj$output.type == "cloud") type <- "p"
  }
  if (bin.cloud == TRUE && type != "b") 
    stop("plot.variogram: object must be a binned variogram with option bin.cloud=T")
  if (bin.cloud == TRUE && all(is.na(obj$bin.cloud))) 
    stop("plot.variogram: object must be a binned variogram with option bin.cloud=T")
  if (bin.cloud == TRUE && any(!is.na(obj$bin.cloud))) 
    boxplot(obj$bin.cloud, varwidth = TRUE, names.x = as.character(obj$u), 
            xlab = "midpoints of distance class", ylab = paste("variogram values / ", 
                                                    obj$estimator.type, "estimator"), ...)
  else {
    u <- obj$u[obj$u <= max.dist]
    obj$v <- as.matrix(obj$v)
    v <- obj$v[obj$u <= max.dist, 1]
    ymax <- max(obj$v[obj$u <= max.dist, ])
    if (!is.null(envelope.obj)) 
      ymax <- max(envelope.obj$v.upper, ymax)
    if (scaled) v <- v/obj$var.mark
    if (ylim == "default") 
      plot(u, v, xlim = c(0, max.dist), ylim = c(0, ymax), 
           xlab = "Distance", ylab = "Semi-variance", type = type, ...)
    else
      plot(u, v, xlim = c(0, max.dist), ylim = ylim, xlab = "Distance", 
           ylab = "Semi-variance", type = type, ...)
    if (var.lines) {
      if (scaled) abline(h = 1, lty = 3)
      else abline(h = obj$var.mark, lty = 3)
    }
    if (ncol(obj$v) > 2) {
      lines.f <- function(v){lines(u,v, type=type)} 
      apply(obj$v[,-1],2, lines.f)
    }
    if (!is.null(envelope.obj)) {
      lines(u, envelope.obj$v.lower, lty = 4)
      lines(u, envelope.obj$v.upper, lty = 4)
    }
  }
  return(invisible())
}
"summary.variomodel" <-
  function(obj)
{
  cat("Summary of the parameter estimation\n")
  cat("-----------------------------------\n")
  if(obj$method == "ML")
    method <- "maximum likelihood"
  if(obj$method == "RML")
    method <- "restricted maximum likelihood"
  if(obj$method == "OLS")
    method <- "ordinary least squares"
  if(obj$method == "WLS")
    method <- "weighted least squares"
  cat(paste("Estimation method:", method, "\n"))
  cat("\n")
  cat(paste("Covariance model:", obj$cov.model))
  if(obj$cov.model == "matern" | obj$cov.model == "powered.exponential" |
     obj$cov.model == "cauchy" | obj$cov.model == "gneiting.matern"
     )
    cat(paste(" with kappa = ", obj$kappa))
  if(!is.null(kappa))
    if(obj$cov.model == "matern" & obj$kappa == 0.5)
      cat(" (exponential)")
  cat("\n")
  cat("Covariance parameters:\n")
  print(c(nugget = obj$nugget, sill = obj$cov.pars[1], range = obj$
          cov.pars[2]))
  cat("\n")
  if(obj$method == "ML" | obj$method == "RML") {
    if(obj$lambda != 1) {
      if(obj$info.lambda$fix.lambda == FALSE)
        cat(paste(
                  "Estimated Box-Cox transformation parameter:",
                  nice(obj$lambda)))
      else cat(paste("Fixed Box-Cox transformation parameter:",
                     obj$lambda))
      cat("\n")
      cat("\n")
    }
    cat("mean parameter(s): ")
    if(length(obj$beta) == 1){
      cat(obj$beta)
      cat("\n")
    }
    else{
      cat("\n")
      print(obj$beta)
    }
    cat("\n")
    print(c(log.L = obj$loglik, AIC = obj$AIC, BIC = obj$BIC, 
            n.params = as.integer(obj$npars)))
  }
  return(invisible())
}
"varcov.spatial" <-
  function(coords = NULL, dists.lowertri = NULL, cov.model = c("exponential",
                                                   "matern", "gaussian", "spherical", "circular", "cubic", "wave", 
                                                   "powered.exponential", "cauchy", "gneiting", "gneiting.matern", "pure.nugget"), kappa = NULL, nugget = 0, cov.pars = stop("no cov.pars argument"), inv = FALSE,
           det = FALSE, func.inv = "cholesky", scaled = FALSE, only.decomposition = FALSE, sqrt.inv = FALSE, test.decomp = FALSE)
{
  require(mva)
  if (func.inv == "chol") func.inv <- "cholesky"
  if(is.null(coords) & is.null(dists.lowertri))
    stop("either argument, coords or dists.lowertri must be provided")
  if(!is.null(coords) & !is.null(dists.lowertri))
    stop("only ONE argument, either coords or dists.lowertri must be provided"		)
  cov.model <- match.arg(cov.model)
  if(!is.null(coords))
    n <- nrow(coords)
  if(!is.null(dists.lowertri)){
    n <- 0.5 * (1 + sqrt(1 + 8 * length(dists.lowertri)))
  }
  tausq <- nugget
  if(is.vector(cov.pars))
    sigmasq <- cov.pars[1]
  else sigmasq <- cov.pars[, 1]
  if(!is.null(coords)) {
    dists.lowertri <- as.vector(dist(coords))
  }
  if(round(1e12*min(dists.lowertri)) == 0)
    warning("Two or more pairs of data at coincident (or very close) locations. This can cause crashes in the matrices operations!\n")
  ##
  ## Building the covariance matrix
  ## 
  varcov <- matrix(0, n, n)
  if(scaled == FALSE) {
    if(cov.pars[2] < 1e-12)
      varcov <- diag(n)  * cov.pars[1]
    else{
      covvec <- cov.spatial(obj = dists.lowertri, cov.model = 
                            cov.model, kappa = kappa, cov.pars = cov.pars)
      varcov[lower.tri(varcov)] <- covvec
      varcov <- t(varcov)
      varcov[lower.tri(varcov)] <- covvec
      diag(varcov) <- tausq + sum(sigmasq)
    }
  }
  else {
    if(cov.pars[2] < 1e-12)
      varcov <- diag(n)
    else{
      if(is.vector(cov.pars))
        cov.pars.sc <- c(1, cov.pars[2])
      else cov.pars.sc <- cbind(1, cov.pars[, 2])
      covvec <- cov.spatial(obj = dists.lowertri, cov.model = 
                            cov.model, kappa = kappa, cov.pars = cov.pars.sc)
      varcov[lower.tri(varcov)] <- covvec
      varcov <- t(varcov)
      varcov[lower.tri(varcov)] <- covvec
      diag(varcov) <- 1 + (tausq/sum(sigmasq))
    }
  }
  ##
  ## Computing decompositions, inverses and determinants
  ## 
  if(test.decomp == TRUE & func.inv == "cholesky"){
    varcov.eig <- eigen(varcov, symmetric = TRUE)
    if(any(varcov.eig$val < 1e-10))
      func.inv <- "eigen"
  }
  if(inv == TRUE |  det == TRUE | only.decomposition == TRUE | sqrt.inv == TRUE) {
    if(func.inv == "cholesky") {
      varcov.sqrt <- chol(varcov)
      if(only.decomposition == FALSE) {
        if(inv == TRUE) {
          invcov <- chol2inv(varcov.sqrt)
        }
        if(det == TRUE)
          cov.logdeth <- sum(log(diag(varcov.sqrt
                                      )))
        if(sqrt.inv == TRUE)
          inverse.sqrt <- solve(varcov.sqrt)
      }
    }
    if(func.inv == "svd") {
      if(test.decomp == FALSE)
      varcov.svd <- svd(varcov, nv = 0)
      if(only.decomposition == TRUE)
        varcov.sqrt <- t(varcov.svd$u %*% (t(varcov.svd$
                                             u) * sqrt(varcov.svd$d)))
      if(inv == TRUE)
        invcov <- t(varcov.svd$u %*% (t(varcov.svd$
                                        u) * (1/varcov.svd$d)))
      cov.logdeth <- sum(log(sqrt(varcov.svd$d)))
      if(sqrt.inv == TRUE)
        inverse.sqrt <- t(varcov.svd$u %*% (t(varcov.svd$
                                              u) * (1/sqrt(varcov.svd$d))))
    }
    if(func.inv == "eigen") {
      varcov.eig <- eigen(varcov, symmetric = TRUE)
      if(only.decomposition == TRUE)
        varcov.sqrt <- (varcov.eig$vec %*% diag(sqrt(
                                                     varcov.eig$val)) %*% t(varcov.eig$
                                                                            vec))
      if(inv == TRUE)
        invcov <- (varcov.eig$vec %*% diag(1/varcov.eig$
                                           val) %*% t(varcov.eig$vec))
      if(det == TRUE)
        cov.logdeth <- sum(log(sqrt(varcov.eig$val)))
      if(sqrt.inv == TRUE)
        inverse.sqrt <- (varcov.eig$vec %*% diag(1/sqrt(varcov.eig$
                                                        val)) %*% t(varcov.eig$vec))
    }
    if(func.inv == "solve") {
      if(det == TRUE)
        stop("this does not compute determinants. Use func.inv = chol, svd or eigen"
             )
      invcov <- solve(varcov)
    }
  }
  ##
  ## Preparing output
  ##
  if(only.decomposition == FALSE) {
    if(det == TRUE){
      if (inv==T){
        result <- list(inverse = invcov, log.det.to.half = 
                       cov.logdeth)
      }
      else{
        result <- list(varcov = varcov, log.det.to.half = 
                       cov.logdeth)
      }
      if(sqrt.inv == TRUE)
        result$sqrt.inverse <- inverse.sqrt
    }
    else{
      if(inv == TRUE)
        result <- invcov
      else
        result <- varcov
    }
  }
  else result <- t(varcov.sqrt)
  return(result)
}
"wireframe.grf" <-
  function (obj, sim.number = 1, ...) 
{
  obj$data <- as.matrix(obj$data)
  wireframe(obj$data[, sim.number] ~ obj$coords[, 1] * obj$coords[, 
                                                                  2], xlab = "Coord X", ylab = "Coord Y", zlab = "Simulated", 
            ...)
}

"wlsfit" <-
function (vario, ini, cov.model = c("exponential", "matern", 
    "gaussian", "spherical", "circular", "cubic", "wave", "powered.exponential", "cauchy", 
    "gneiting", "gneiting.matern", "pure.nugget"), kappa = NULL, simul.number = NULL, 
    max.dist = "all", minimisation.function = "nlm", lower = 0, fix.nugget = FALSE, 
    nugget = 0, weight = c("npairs", "cressie"), print.level=0, messages.screen=TRUE) 
{
  call <- match.call()
  weight <- match.arg(weight)
  if (minimisation.function != "nlm") 
    stop("minimization function available is: nlm")
  if (is.matrix(vario$v) & is.null(simul.number)) 
    stop("object in vario$v is a matrix. This function works for only 1 empirical variogram")
  if (!is.null(simul.number)) 
    vario$v <- vario$v[, simul.number]
  if(is.matrix(ini)) {
    inilength <- dim(ini)[2]
    if(fix.nugget == FALSE & inilength != 3)
      stop("wrong number of columns for ini (must be 3)")
    if(fix.nugget == TRUE & inilength != 2)
      stop("wrong number of columns for ini (must be 2)")
  }
  else {
    inilength <- length(ini)
    if (fix.nugget == FALSE & inilength != 3) 
      stop("wrong length for ini (must be 3)")
    if (fix.nugget == TRUE & inilength != 2) 
      stop("wrong length for ini (must be 2)")
  }
  cov.model <- match.arg(cov.model)
  if (max.dist == "all") 
    data <- list(u = vario$u, v = vario$v, n = vario$n)
  else data <- list(u = vario$u[vario$u <= max.dist], v = vario$v[vario$u <= 
                                                        max.dist], n = vario$n[vario$u <= max.dist])
  holdit <- data$v
  holdit[is.na(holdit)] <- 0
  data$v <- holdit
  data$kappa <- kappa
  data$cov.model <- cov.model
  .global.list <<- list(data = data, fix.nugget = fix.nugget,
                        nugget = nugget, 
                        kappa = kappa, cov.model = cov.model,
                        weight = weight)
  .global.lower <<- lower
  if(is.matrix(ini) | is.data.frame(ini)) {
    ini <- as.matrix(ini)
    cat("wlsfit: searching for the best initial value\n")
    ini.search <- ini
    dimnames(ini.search) <- list(NULL, NULL)
    wls.ini <- round(100000000. * apply(ini.search, 1, loss.wlsvario))
    ini <- as.vector(ini.search[wls.ini == min(wls.ini),])
    cat("wlsfit: best initial value:\n")
    if(fix.nugget == TRUE){
      names(ini) <- c("sill", "range")
      print(ini)
    }
    else{
      names(ini) <- c("nugget", "sill", "range")
      print(ini)
    }
  }
  result <- nlm(loss.wlsvario, ini, iterlim = 10000,print.level=print.level)
  if (exists(".temp.theta", w=1)){
    result$estimate <- .temp.theta
    remove(".temp.theta", pos=1, inherits = TRUE)
  }
  rm(.global.list, .global.lower, inherits = TRUE)
  value <- result$minimum
  message <- paste("Exit code:", result$code)
  if (fix.nugget == FALSE) {
    nugget <- result$estimate[1]
    cov.pars <- as.vector(result$estimate[2:3])
  }
  else {
    cov.pars <- as.vector(result$estimate[1:2])
  }
  estimation <- list(nugget = nugget, cov.pars = cov.pars, 
                     cov.model = cov.model, kappa = kappa, value = value, 
                     max.dist = max(data$u), minimisation.function = minimisation.function, message = message)
  estimation$method <- "WLS"
  estimation$call <- call
  class(estimation) <- "variomodel"
  if(messages.screen){
    cat("wlsfit: estimated model parameters are:\n")
    cat(paste("covariance model:", cov.model))
    if(cov.model == "matern" | cov.model == "powered.exponential" | 
       cov.model == "cauchy" | cov.model == "gneiting.matern")
      cat(paste(" with kappa =", kappa))
    if(!is.null(kappa))
      if(cov.model == "matern" & kappa == 0.5)
        cat(" (exponential)")
    cat("\n")
    print(c(nugget=estimation$nugget, sill=estimation$cov.pars[1], range=estimation$cov.pars[2]))
  }
  return(estimation)
}
"xyzplot" <- 
  function(geodata, coords=geodata$coords, data=geodata$data, eyes, planepts = FALSE, ...)
{
  xr <- range(coords[, 1])
  yr <- range(coords[, 2])
  zr <- range(data)
  if(missing(eyes))
    eyes <- c(8, 8, 3)
  eyex <- min(xr) - eyes[1] * (xr[2] - xr[1])
  eyey <- min(yr) - eyes[2] * (yr[2] - yr[1])
  eyez <- min(zr) + 2 * eyes[3] * (zr[2] - zr[1])
  eye <- c(eyex, eyey, eyez)
### eye_c(-113.6,29,20000)
  pp <- persp(xr, yr, matrix(zr[1], ncol = 2, nrow = 2), zlim = zr, eye
              = eye, box = FALSE, xlab = "Coord X", ylab = "Coord Y", zlab = 
              "data")
  ptsin3d <- perspp(coords[, 1], coords[, 2], data, pp)
  ptsonplane <- perspp(coords[, 1], coords[, 2], rep(zr[1], length(data)), pp)
  points(ptsin3d, pch = 16, ...)
  if(planepts) {
    points(ptsonplane)
  }
  segments(ptsin3d$x, ptsin3d$y, ptsonplane$x, ptsonplane$y)
}
"trend.spatial" <-
  function(trend, coords=NULL)
{
  if(inherits(trend, "formula")){
    trend.mat <- model.matrix(trend)
  }
  else{
    if(trend == "cte")
      trend.mat <- as.matrix(rep(1, nrow(coords)))
    else
      if(trend == "1st")
        trend.mat <- cbind(1, coords)
      else
        if(trend == "2nd")
          trend.mat <- cbind(1, coords, coords[, 1]^2, coords[
                                                            , 2]^2, coords[, 1] * coords[, 2])
	else
          stop("external trend must be provided for data locations to be estimated using the argments `trend.d` and `trend.l`. Both (trend.d and trend.l) must be the same `cte`, `1st`, `2nd` or given by a formula of the type ~X where X is a matrix or vector of covariates."
               )
  }
  trend.mat <- as.matrix(trend.mat)
  dimnames(trend.mat) <- list(NULL, NULL)
  return(trend.mat)
}
  
"krige.conv" <-
function (geodata, coords=geodata$coords, data=geodata$data, locations, trend.d = "cte", trend.l = "cte", 
            type.krige = "ok", beta = NULL, cov.model = c("exponential", 
                                              "matern", "gaussian", "spherical", "circular", "cubic", "wave", "powered.exponential", 
                                              "cauchy", "gneiting", "gneiting.matern", "pure.nugget"), kappa = 0.5, 
            cov.pars = stop("covariance parameters (sigmasq and phi) needed"), 
            nugget = 0, micro.scale = 0, lambda = 1, n.samples.backtransform = 500, aniso.pars = FALSE,
            signal = FALSE, messages.screen = TRUE, dist.epsilon = 1e-10,
            n.sim, ...) 
{
  call <- match.call()
  if(missing(n.sim))
    n.sim <- as.integer(0)
  if(lambda != 1) {
    if(messages.screen)
      print("Data transformation (Box-Cox) performed.")
    if(lambda == 0)
      data <- log(data)
    else data <- ((data^lambda) - 1)/lambda
  }
  cov.model <- match.arg(cov.model)
  if(micro.scale > nugget)
    stop("micro.scale must be in the interval [0, nugget]")
  if (type.krige == "ok" | type.krige == "OK") 
    beta.prior <- "flat"
  if (type.krige == "sk" | type.krige == "SK"){
    if(missing(beta)) stop("argument beta must be provided in order to perform simple kriging")
    beta.prior <- "deg"
  }
  if (is.vector(coords)) 
    coords <- as.matrix(cbind(coords, 0))
  else coords <- as.matrix(coords)
  if (is.vector(locations)) {
    if (length(locations) == 2) {
      locations <- t(as.matrix(locations))
      if (messages.screen) 
        print("WARNING: THE FUNCTION IS CONSIDERING THAT YOU HAVE ENTERED WITH 1 POINT IN TWO DIMENSIONS TO BE PREDICT")
    }
    else locations <- as.matrix(cbind(locations, 0))
  }
  else locations <- as.matrix(locations)  
  if(inherits(trend.d, "formula") | inherits(trend.l, "formula")){
    if(any(locations != "no"))
      if((inherits(trend.d, "formula") == FALSE) | (inherits(trend.l, "formula") == FALSE))
        stop("trend.d and trend.l must have similar specification")
    if(messages.screen)
      print("Kriging with external trend to be performed using covariates provided by the user")
  }
  else{
    if (any(locations != "no") & (trend.d != trend.l)){
      stop("trend.l is different from trend.d")
    }
    if(messages.screen){
      if(trend.d == "cte" & type.krige == "sk")
        print("Simple kriging to be performed with constant mean provided by the user"
              )
      if(trend.d == "cte" & type.krige == "ok")
        print("Ordinary kriging to be performed filtering a constant mean")
      if(trend.d == "1st")
        print("Trend (or universal) kriging to be performed filtering a 1st degree polinomial")
      if(trend.d == "2nd") 
        print("Trend (or universal) kriging to be performed filtering a 2nd degree polinomial")
    }
  }
  trend.d <- trend.spatial(trend=trend.d, coords=coords)
  trend.l <- trend.spatial(trend=trend.l, coords=locations)
  dimnames(coords) <- list(NULL, NULL)
  dimnames(locations) <- list(NULL, NULL)
  if (any(aniso.pars) == TRUE) {
    coords <- coords.aniso(coord = coords, aniso.pars = aniso.pars)
    locations <- coords.aniso(coord = locations, aniso.pars = aniso.pars)
  }
  n <- length(data)
  ni <- length(trend.l)
  tausq <- nugget
  if (is.vector(cov.pars)) {
    sigmasq <- cov.pars[1]
    phi <- cov.pars[2]
  }
  else {
    sigmasq <- cov.pars[, 1]
    phi <- cov.pars[, 2]
  }
  sill.partial <- sum(sigmasq)
  sill.tot <- tausq + sill.partial
  invcov <- varcov.spatial(coords = coords, cov.model = cov.model, 
                           kappa = kappa, nugget = nugget, cov.pars = cov.pars, 
                           inv = TRUE)
  ttrend.d <- t(trend.d)
  ttiv <- ttrend.d %*% invcov
  ttivtt <- ttiv %*% trend.d
  ittivtt <- solve(ttivtt)
  ttivz <- ttiv %*% data
  if (beta.prior == "flat") 
    beta.flat <- ittivtt %*% ttivz
  v0mat <- apply(locations, 1, dist0.krige, coord = coords)
  if(n.sim ==0) remove("locations")
  if(micro.scale != 0)
    v0mat <- ifelse(v0mat < dist.epsilon, micro.scale + sill.partial,
                    cov.spatial(obj = v0mat, cov.model = cov.model, 
                                kappa = kappa, cov.pars = cov.pars))
  else
    v0mat <- cov.spatial(obj = v0mat, cov.model = cov.model, 
                         kappa = kappa, cov.pars = cov.pars)
  tv0iv <- t(apply(v0mat, 2, crossprod, y = invcov))
remove("invcov")	
  tv0ivz <- tv0iv %*% data
  tv0ivx <- apply(tv0iv, 1, crossprod, y = trend.d)
  if(n.sim > 0) {
    reduce.var <- crossprod(v0iv, v0mat)
    tv0ivv0 <- diag(reduce.var)
  }
  else
    tv0ivv0 <- diag(tv0iv %*% v0mat)
  remove("v0mat")
  tb <- t(trend.l) - tv0ivx
  remove("trend.l")
  b <- t(tb)
  if (beta.prior == "deg") {
    predict <- tv0ivz + b %*% beta
    if (signal == TRUE) 
      krige.var <- sill.partial - tv0ivv0
    else krige.var <- sill.tot - tv0ivv0
    beta.est <- cat("SK : beta provided by user")
  }
  if (beta.prior == "flat") {
    predict <- tv0ivz + b %*% beta.flat
    bi <- apply(tb, 2, crossprod, y = ittivtt)
    if (ncol(trend.d) > 1) 
      bi <- t(bi)
    if(n.sim > 0) {
      ok.add.var <- (bi %*% tb)
      reduce.var <- reduce.var + ok.add.var
      bitb <- diag(ok.add.var)
    }
    else
      bitb <- diag(bi %*% tb)
    if (signal == TRUE) 
      krige.var <- sill.partial - tv0ivv0 + bitb
    else krige.var <- sill.tot - tv0ivv0 + bitb
    beta.est <- beta.flat
  }
  predict <- as.vector(predict)
  krige.var <- as.vector(round(krige.var, dig = 8))
  if(any(krige.var < 0))
    print("Negative kriging variance found! Investigate why this is happening."
          )
  message <- "Kriging performed in global neighbourhood"
  if(messages.screen)
    print(message)
  if(lambda != 1) {
    if(messages.screen)
      print("Back-transforming the predictions according to the (Box-Cox) parameter lambda")
    if(lambda == 0) {
      predict.transf <- predict
      predict <- exp(predict.transf) - 0.5 * krige.var
      krige.var <- (exp(2 * predict.transf - krige.var)) *
        (exp(krige.var) - 1)
    }
    if(lambda > 0) {
      if(messages.screen)
        print("Back-transformation done by sampling from the resulting (normal) predictive distribution"
              )
      ap.warn <- options()$warn
      options(warn = -1)
      temp.data <- matrix(rnorm(ni * n.samples.backtransform,
                                mean = predict, sd = sqrt(krige.var)), nrow = 
                          ni)
      options(warn = ap.warn)
      temp.data[(krige.var == 0),  ] <- predict[(krige.var ==
                   0)]
      temp.data[temp.data < -1/lambda] <- -1/lambda
      temp.data <- ((temp.data * lambda) + 1)^(1/lambda)
###      temp.data[is.na(temp.data)] <- Inf
      predict <- as.vector(apply(temp.data, 1, mean))
      krige.var <- as.vector(apply(temp.data, 1, var))
    }
    if(lambda < 0) {
      print("Resulting distribution has no mean for lambda < 0 - back transformation not performed"
            )
    }
  }
  results <- list(predict = predict, krige.var = krige.var, beta.est = 
                  beta.est, message = message, call = call)
############## Sampling from the resulting distribution #####################
  if(n.sim > 0) {
    print("sampling realisations from the resulting distribution")
    varcov <- varcov.spatial(coords = locations,
                             cov.model = cov.model,
                             cov.pars = cov.pars,
                             kappa = kappa, nugget = nugget) - reduce.var
    results$samples <-  predict + (chol(varcov) %*% matrix(rnorm(ni * n.sim), ncol=n.sim))
  }
#######################################
  class(results) <- "kriging"
  return(results)
}

"lines.grid.bayes" <- 
function(obj, max.dist, length = 100, summary.posterior = c("mode", "median",
	"mean"), ...)
{
	spost <- match.arg(summary.posterior)
	if(is.null(obj$call$cov.model))
		cov.model <- "exponential"
	else {
		cov.model <- obj$call$cov.model
		if(obj$call$cov.model == "matern" | obj$call$cov.model == 
			"powered.exponential" | obj$call$cov.model == "cauchy" |
			obj$call$cov.model == "gneiting-matern")
			kappa <- obj$call$kappa
		else kappa <- NULL
	}
	distance <- seq(0, max.dist, length = length)
	cov.pars <- c(obj$sigmasq.summary[spost], obj$phi.summary[spost])
	names(cov.pars) <- NULL
	if(any(obj$tausq.summary != 0))
		nugget <- obj$tausq.summary[spost] * cov.pars[1]
	else nugget <- 0
	names(nugget) <- NULL
	sill.tot <- nugget + cov.pars[1]
	gamma <- (nugget + cov.pars[1]) - cov.spatial(distance, cov.model = 
		cov.model, kappa = kappa, cov.pars = cov.pars)
	lines(distance, gamma, ...)
}

"image.kriging" <-
  function (kriging.obj, locations = kriging.obj$locations, values = kriging.obj$predict, ...) 
{
  par.ori <- par(no.readonly = TRUE)
  on.exit(par(par.ori))
  if(is.null(locations))
    stop("prediction locations must be provided")
  if(is.null(values))
    stop("data.to be plotted must be provided")
  locations <- locations[order(locations[, 2], locations[,1]), ]
  x <- as.numeric(levels(as.factor(locations[, 1])))
  nx <- length(x)
  y <- as.numeric(levels(as.factor(locations[, 2])))
  ny <- length(y)
  if (nx == ny) 
    par(pty = "s")
  values <- matrix(values, ncol = ny)
  image(x, y, values, ...)
    return(invisible())
}
"rfm.bin" <-
function (cloud, l = 15, uvec = "default", nugget.tolerance = 0, 
    estimator.type = c("classical", "robust"), bin.cloud = FALSE) 
{
    if (all(uvec == "default")) 
        uvec <- seq(0, max(cloud$u), l = l)
    estimator.type <- match.arg(estimator.type)
    if (nugget.tolerance > 0) {
        dnug <- mean(cloud$u[cloud$u <= nugget.tolerance])
        cloud$u[cloud$u <= nugget.tolerance] <- 0
        uvec <- uvec[uvec > nugget.tolerance]
    }
    u <- c(0, uvec)
    n <- length(u)
    min.dist <- min(cloud$u)
    d <- 0.5 * (u[3:n] - u[2:(n - 1)])
    low <- c(0, (u[2:(n - 1)] + d))
    high <- c((u[3:n] - d), (d[n - 2] + u[n]))
    nc <- n - 1
    if (!is.matrix(cloud$v)) {
        vbin <- rep(0, nc)
        nbin <- rep(0, nc)
        sdbin <- rep(0, nc)
        if (bin.cloud == TRUE) 
            bins.clouds <- list()
        for (i in 1:nc) {
            ind <- (cloud$u > low[i]) & (cloud$u <= high[i])
            vbin[i] <- mean(cloud$v[ind])
            if (bin.cloud == TRUE) 
                bins.clouds[[i]] <- cloud$v[ind]
            nbin[i] <- sum(ind)
            if (estimator.type == "robust") 
                vbin[i] <- ((vbin[i])^4)/(0.914 + (0.988/nbin[i]))
            if (nbin[i] > 0) 
                sdbin[i] <- sqrt(var(cloud$v[ind]))
            else sdbin[i] <- NA
            NULL
        }
        if (uvec[1] == 0) 
            uvec[1] <- (low[1] + high[1])/2
        if (min.dist == 0) {
            ind <- (cloud$u == 0)
            n.zero <- sum(ind)
            v.zero <- mean(cloud$v[ind])
            if (bin.cloud == TRUE) {
                bins.clouds[[2:(length(bins.clouds) + 1)]] <- bins.clouds[1:n]
                bins.clouds[[1]] <- cloud$v[ind]
            }
            if (estimator.type == "robust") 
                v.zero <- ((v.zero)^4)/(0.914 + (0.988/n.zero))
            if (n.zero > 0) 
                sd.zero <- sqrt(var(cloud$v[ind]))
            else sd.zero <- NA
            uvec <- c(0, uvec)
            vbin <- c(v.zero, vbin)
            nbin <- c(n.zero, nbin)
            sdbin <- c(sd.zero, sdbin)
        }
        u <- uvec[!is.na(vbin)]
        v <- vbin[!is.na(vbin)]
        n <- nbin[!is.na(vbin)]
        sd <- sdbin[!is.na(vbin)]
        if (bin.cloud == TRUE) 
            bins.clouds <- bins.clouds[!is.na(vbin)]
    }
    else {
        if (bin.cloud == TRUE) 
            stop("option bins.cloud=T allowed only for 1 variable")
        nvcols <- ncol(cloud$v)
        vbin <- matrix(0, nrow = nc, ncol = nvcols)
        nbin <- rep(0, nc)
        sdbin <- matrix(0, nrow = nc, ncol = nvcols)
        for (i in 1:nc) {
            ind <- (cloud$u >= low[i]) & (cloud$u < high[i])
            nbin[i] <- sum(ind)
            for (j in 1:nvcols) {
                vbin[i, j] <- mean(cloud$v[ind, j])
                if (estimator.type == "robust") 
                  vbin[i, j] <- ((vbin[i, j])^4)/(0.914 + (0.988/nbin[i]))
                if (nbin[i] > 0) 
                  sdbin[i, j] <- sqrt(var(cloud$v[ind, j]))
                else sdbin[i, j] <- NA
            }
            NULL
        }
        if (uvec[1] == 0) 
            uvec[1] <- (low[1] + high[1])/2
        if (min.dist == 0) {
            v.zero <- rep(0, nvcols)
            n.zero <- rep(0, nvcols)
            sd.zero <- rep(0, nvcols)
            for (j in 1:nvcols) {
                ind <- (cloud$u == 0)
                n.zero[j] <- sum(ind)
                v.zero[j] <- mean(cloud$v[ind, j])
                if (estimator.type == "robust") 
                  v.zero[j] <- ((v.zero[j])^4)/(0.914 + (0.988/n.zero[j]))
                if (n.zero[j] > 0) 
                  sd.zero[j] <- sqrt(var(cloud$v[ind, j]))
                else sd.zero[j] <- NA
                uvec <- c(0, uvec)
                vbin <- rbind(v.zero, vbin)
                nbin <- c(n.zero, nbin)
                sdbin <- rbind(sd.zero, sdbin)
            }
        }
        u <- uvec[!is.na(vbin[, 1])]
        n <- nbin[!is.na(vbin[, 1])]
        v <- matrix(0, nrow = length(u), ncol = nvcols)
        sd <- matrix(0, nrow = length(u), ncol = nvcols)
        for (j in 1:nvcols) {
            v[, j] <- vbin[!is.na(vbin[, j]), j]
            sd[, j] <- sdbin[!is.na(vbin[, j]), j]
        }
    }
    if (nugget.tolerance > 0) {
        u[1] <- nugget.tolerance
    }
    result <- list(u = u, v = v, n = n, sd = sd, output.type = "bin", bins.lim = c(low[1],high))
    if (!is.matrix(cloud$v) && bin.cloud == TRUE) 
        result$bin.cloud <- bins.clouds
    if (!is.null(class(cloud))) 
        class(result) <- class(cloud)
    return(result)
}
"points.geodata" <- 
  function(geodata, coords = geodata$coords, data = geodata$data, col.data = 1,
           pt.sizes = c("data.proportional", "rank.proportional", "quintiles", "quartiles", "deciles", 
             "equal"), cex.min, cex.max, pch.seq, col.seq, add.to.plot = FALSE,  round.quantiles = FALSE, graph.pars = TRUE,...)
{
  pt.sizes <- match.arg(pt.sizes)
  if(add.to.plot == FALSE) {
    plot(apply(coords, 2, range), type = "n")
  }
  if(missing(cex.min))
    cex.min <- 0.5
  if(missing(cex.max))
    cex.max <- 1.5
  if(is.matrix(data))
    data <- as.vector(data[, col.data])
  graph.list <- list()
  if(pt.sizes == "quintiles" | pt.sizes == "quartiles" | pt.sizes == "deciles") {
    r.col <- rainbow(13)
    if(pt.sizes == "quintiles") {
      n.quant <- 5
      if(missing(col.seq))
        col.seq <- r.col[c(10,7,5,3,1)]
    }
    if(pt.sizes == "quartiles") {
      n.quant <- 4
      if(missing(col.seq))
        col.seq <- r.col[c(10,5,3,1)]
    }
    if(pt.sizes == "deciles") {
      n.quant <- 10
      if(missing(col.seq))
        col.seq <- r.col[10:1]
    }
    if(missing(pch.seq))
      pch.seq <- rep(21, n.quant)
    cex.pt <- seq(cex.min, cex.max, l = n.quant)
    data.quantile <- quantile(data, probs = seq(0, 1, by = 
                                      (1/n.quant)))
    if (round.quantiles == TRUE){
      data.quantile[1] <- floor(data.quantile[1])
      data.quantile[n.quant+1] <- ceiling(data.quantile[n.quant+1])
      data.quantile <- round(data.quantile)
    }
    graph.list$quantiles <- data.quantile
    graph.list$cex <- cex.pt
    graph.list$col <- col.seq
    graph.list$pch <- pch.seq
    data.quantile[1] <- 0.9 * data.quantile[1]
    for(i in 1:n.quant){
      points(coords[((data > data.quantile[i]) & (data <= 
                                                  data.quantile[i + 1])),  ], pch = 21
             , cex = cex.pt[i], bg = col.seq[i], ...)        
    }
  }
  if(pt.sizes == "rank.proportional") {
    if(missing(pch.seq))
      pch.seq <- 21
    if(missing(col.seq))
      col.seq <- 0
    data.quantile <- range(data)
    n <- length(data)
    coords.order <- coords[order(data),  ]
    data.order <- data[order(data)]
    pt.size <- seq(cex.min, cex.max, l = n)
    graph.list$cex <- range(pt.size)
    graph.list$pch <- unique(range(pch.seq))
	for(i in 1:n)
      points(coords.order[i,  , drop = FALSE], cex = pt.size[i], 
	pch = pch.seq, bg = col.seq[i],	...)
  }	
  if(pt.sizes == "data.proportional") {
    if(missing(pch.seq))
      pch.seq <- 21
    if(missing(col.seq))
      col.seq <- 0
    r.y <- range(data)
    n <- length(data)
    size <- cex.min + ((data - r.y[1]) * (cex.max - cex.min))/(r.y[2] - r.y[1])
    graph.list$cex <- c(cex.min, cex.max)
    graph.list$pch <- unique(range(pch.seq))
    for(i in 1:n)
      points(coords[i,  , drop = FALSE], cex = size[i], 
	pch = pch.seq, bg = col.seq[i],	...)
  }	
  if(pt.sizes == "equal"){
    if(missing(pch.seq)) 
      pch.seq <- 16
    points(coords, pch = pch.seq, ...)
  }
  if (graph.pars == TRUE)
    return(graph.list)
  else
    return(invisible())
}
"my.ps" <- 
function(file = "junk", height = 2.5, width = 5, labels = TRUE, cex = 
	0.69999999999999996)
{
	ps.options(horizontal = FALSE)
	ps.options(width = width, height = height)
	postscript(file = file, print.it = FALSE)
	if(labels == FALSE)
		par(mar = c(1.8, 1.8, 0, 0), oma = c(0, 0, 1.45, 0), cex = cex,
			las = 1, mgp = c(1.5, 0.5, 0))
	else par(mar = c(2.7000000000000002, 2.7000000000000002, 0, 
			0.10000000000000001), oma = c(0, 0, 1.45, 0), cex = cex,
			las = 1, mgp = c(2, 0.35, 0))
	return(invisible())
}

"persp.grf" <- 
function(obj, sim.number = 1, ...)
{
	x <- as.numeric(levels(as.factor(obj$coords[, 1])))
	nx <- length(x)
	y <- as.numeric(levels(as.factor(obj$coords[, 2])))
	ny <- length(y)
	obj$data <- as.matrix(obj$data)
	n <- nrow(obj$data)
	if(nx * ny != n)
		stop("Probably irregular grid")
	m <- matrix(obj$data[, sim.number], ncol = ny)
	persp(x, y, m, ...)
	return(invisible())
}

"coords.aniso" <- 
  function(coords, aniso.pars)
{
  n <- nrow(coords)
  rotation <- aniso.pars[1]
  lambda <- aniso.pars[2]
  if(lambda < 1)
    stop("anisotropy ratio must be geater then 1")
  rm <- matrix(c(cos(rotation),  - sin(rotation), sin(rotation), cos(
                                                                     rotation)), ncol = 2)
  tm <- matrix(c(1, 0, 0, 1/lambda), ncol = 2)
  coords.cor <- coords %*% rm %*% tm
  return(coords.cor)
}

"variog.env" <-
  function (obj.variog, coords, model.pars, nsim = 99, messages.screen = TRUE) 
{
  call <- match.call()
  ##
  ## reading input
  ##
  if(!is.null(model.pars$beta)) beta <- model.pars$beta
  else beta <- 0
  if(!is.null(model.pars$cov.model))
    cov.model <- model.pars$cov.model
  else cov.model <- "exponential"
  if(!is.null(model.pars$kappa)) kappa <- model.pars$kappa
  else kappa <- 0.5
  if(!is.null(model.pars$nugget)) nugget <- model.pars$nugget
  else nugget <- 0
  cov.pars <- model.pars$cov.pars
  if(!is.null(obj.variog$estimator.type))
    estimator.type <- obj.variog$estimator.type
  else estimator.type <- "classical"
  if (obj.variog$output.type != "bin") 
    stop("envelops can be computed only for binned variogram")
  ##
  ## generating simulations from the model with parameters provided
  ##
  if (messages.screen) 
    cat(paste("variog.env: generating", nsim, "simulations (with ",
              obj.variog$n.data, 
              "points each) using the function grf\n"))
  simula <- grf(obj.variog$n.data, grid = as.matrix(coords),
                cov.model = cov.model, cov.pars = cov.pars,
                nugget = nugget, kappa = kappa, nsim = nsim,
                messages.screen = FALSE, lambda = obj.variog$lambda)
  if(messages.screen)
    cat("variog.env: adding the mean or trend\n")
  x.mat <- trend.spatial(trend=obj.variog$trend, coords=coords)
  simula$data <- as.vector(x.mat %*% beta) + simula$data
  ##
  ## computing empirical variograms for the simulations
  ##
  if (messages.screen) 
    cat(paste("variog.env: computing the empirical variogram for the", 
              nsim, "simulations\n"))
  nbins <- length(obj.variog$bins.lim) - 1
  bin.f <- function(sim){
    cbin <- vbin <- sdbin <- rep(0, nbins)  
    temp <- .C("binit",
               as.integer(obj.variog$n.data),
               as.double(as.vector(coords[,1])),
               as.double(as.vector(coords[,2])),
               as.double(as.vector(sim)),
               as.integer(nbins),
               as.double(as.vector(obj.variog$bins.lim)),
               as.integer(estimator.type == "robust"),
               as.double(max(obj.variog$u)),
               as.double(cbin),
               vbin = as.double(vbin),
               as.integer(FALSE),
               as.double(sdbin)
               )$vbin
    return(temp)
  }
  simula.bins <- apply(simula$data, 2, bin.f)
  simula.bins <- simula.bins[obj.variog$ind.bin,]
  ##
  ## computing envelops
  ##
  if (messages.screen) 
    cat("variog.env: computing the envelops\n")
  limits <- apply(simula.bins, 1, range)
  res.env <- list(u = obj.variog$u, v.lower = limits[1, ],
                  v.upper = limits[2,])
  return(res.env)
}

"variog" <-
  function(geodata, coords=geodata$coords, data=geodata$data, 
           uvec = "default", trend = "cte", lambda = 1, 
           option = c("bin", "cloud", "smooth"), 
           estimator.type = c("classical", "robust"), 
           nugget.tolerance = 0, max.dist = NULL, pairs.min = 2, 
           bin.cloud = FALSE, ...) 
{
  require(mva)
  require(modreg)
  ##
  ## reading and checking input
  ##
  call <- match.call()
  coords <- as.matrix(coords)
  data <- as.matrix(data)
  data.var <- apply(data, 2, var)
  n.data <- nrow(coords)
  n.datasets <- ncol(data)
  if(ncol(data) == 1)
    data <- as.vector(data)
  option <- match.arg(option)
  estimator.type <- match.arg(estimator.type)
  ##
  ## Box-Cox transformation
  ##
  if (lambda != 1){
    if (lambda == 0) data <- log(data)
    else data <- ((data^lambda) - 1)/lambda
  }
  ##
  ## Removing OLS trend
  ##
  xmat <- trend.spatial(trend=trend, coords=coords)
  if(trend != "cte"){
    if(is.vector(data)){
      data <- lm(data ~ xmat + 0)$residuals
      names(data) <- NULL
    }
    else{
      only.res <- function(y, x){lm(y ~ x + 0)$residuals}
      data <- apply(data,2,only.res, x=xmat)
    }
  }
  u <- as.vector(dist(as.matrix(coords)))
  ##
  if(option == "bin" & bin.cloud == FALSE){
    if(!is.null(max.dist))
      umax <- max(u[u < max.dist])
    else umax <- max(u)
      if (all(uvec == "default")) 
      uvec <- seq(0, umax, l = 15)
    ubin <- c(0, uvec)
    nvec <- length(ubin)
    d <- 0.5 * diff(ubin[2:nvec])
    bins.lim <- c(0, (ubin[2:(nvec - 1)] + d),(d[nvec - 2] + ubin[nvec]))
    if (uvec[1] == 0 & nugget.tolerance == 0) 
      uvec[1] <- (bins.lim[1] + bins.lim[2])/2
    if (nugget.tolerance > 0){
      bins.lim <- c(0, nugget.tolerance, bins.lim[bins.lim > nugget.tolerance])
      uvec <- c(0, (bins.lim[-(1:2)] - 0.5*diff(bins.lim)[-1]))
    }
    nbins <- length(bins.lim) - 1
    if(is.null(max.dist)) max.dist <- max(bins.lim)
    bin.f <- function(data){
      cbin <- vbin <- sdbin <- rep(0,nbins)
      result <- .C("binit",
                   as.integer(n.data),
                   as.double(as.vector(coords[,1])),
                   as.double(as.vector(coords[,2])),
                   as.double(as.vector(data)),
                   as.integer(nbins),
                   as.double(as.vector(bins.lim)),
                   as.integer(estimator.type == "robust"),
                   as.double(max.dist),
                   cbin = as.integer(cbin),
                   vbin = as.double(vbin),
                   as.integer(TRUE),
                   sdbin = as.double(sdbin)
                   )[c("vbin", "cbin", "sdbin")]
    }
    result <- array(unlist(lapply(as.data.frame(data),bin.f)), dim=c(nbins, 3, n.datasets))
    indp <- (result[,2,1] >= pairs.min)
    result <- list(u = uvec[indp], v = result[indp,1,], n = result[indp,2,1],
                   sd = result[indp,3,], bins.lim = bins.lim, ind.bin = indp)
  }
  else{
    if (is.matrix(data)) {
      v <- matrix(0, nrow = length(u), ncol = ncol(data))
      for (i in 1:ncol(data)) {
        v[, i] <- as.vector(dist(data[, i]))
        if (estimator.type == "robust") 
          v[, i] <- v[, i]^(0.5)
        else v[, i] <- (v[, i]^2)/2
      }
      if (!is.null(max.dist)) {
        v <- v[u <= max.dist, ]
        u <- u[u <= max.dist]
      }
    }
    else {
      v <- as.vector(dist(data))
      if (estimator.type == "robust") 
        v <- v^(0.5)
      else v <- (v^2)/2
      if (is.numeric(max.dist)) {
        v <- v[u <= max.dist]
        u <- u[u <= max.dist]
      }
    }
    if (option == "cloud") {
      result <- list(u = u, v = v)
    }
    if (option == "bin") {
      result <- rfm.bin(cloud = list(u = u, v = v),
                        estimator.type = estimator.type, 
                        uvec = uvec, nugget.tolerance = nugget.tolerance, 
                        bin.cloud = bin.cloud, ...)
      indp <- rep(TRUE, length(u))
      if (pairs.min > 0) {
        indp <- (result$n >= pairs.min)
        if (is.matrix(result$v)) {
          result$v <- result$v[indp, ]
          result$sd <- result$sd[indp, ]
        }
        else {
          result$v <- result$v[indp]
          result$sd <- result$sd[indp]
        }
        result$u <- result$u[indp]
        result$n <- result$n[indp]
      }
      result$bins.lim <- bin.lims
      result$ind.bin <- indp
    }
    if (option == "smooth") {
      require(modreg)
      if (is.matrix(v)) 
        stop("smooth not yet available for several variables")
      temp <- ksmooth(u, v, ...)
      result <- list(u = temp[[1]], v = temp[[2]])
    }
  }
  result <- c(result, list(var.mark = data.var, output.type = option,
                           estimator.type = estimator.type, n.data = n.data,
                           lambda = lambda, trend = trend, call = call))
  class(result) <- "variogram"
  return(result)
}


"cov.spatial" <-
function(obj, cov.model = c("exponential", "matern", "gaussian", "spherical",
                  "circular", "cubic", "wave", "powered.exponential",
                "cauchy", "gneiting", "gneiting.matern", "pure.nugget"),
         kappa, cov.pars = stop("no cov.pars argument provived"))
{
  ##
  ## checking/reading input
  ##
  cov.model <- match.arg(cov.model)
  if(cov.model == "matern" | cov.model == "powered.exponential" | 
     cov.model == "cauchy" | cov.model == "gneiting.matern"){
    if(missing(kappa))
      stop("for matern, powered.exponential, cauchy, gneiting and gneiting.matern covariance functions the parameter kappa must be provided")
  if(cov.model == "gneiting.matern" & length(kappa) != 2)
    stop("gneiting.matern correlation function model requires a vector with 2 parameters in the argument kappa")
  if((cov.model == "matern" | cov.model == "powered.exponential" | 
      cov.model == "cauchy") & length(kappa) != 1)
    stop("for this choice of  correlation function model kappa should be a scalar parameter")
  }
  if(is.vector(cov.pars))
    sigmasq <- cov.pars[1]
  else sigmasq <- cov.pars[, 1]
  if(is.vector(cov.pars))
    phi <- cov.pars[2]
  else phi <- cov.pars[, 2]
  if(is.vector(cov.pars))
    ns <- 1
  else ns <- nrow(cov.pars)
  covs <- array(0, dim = dim(obj))
  #
  # computing covariances
  #
  for(i in 1:ns) {
    if(phi[i] < 1e-12)
      cov.model <- "pure.nugget"
    cov <- switch(cov.model,
                  pure.nugget = rep(0, length(obj)),
                  wave = (1/obj) * (phi[i] * sin(obj/phi[i])),
                  exponential = exp( - (obj/phi[i])),
                  matern = matern(u = obj, phi = phi[i], kappa = kappa),
                  gaussian = exp( - ((obj/phi[i])^2)),
                  spherical = ifelse(obj < phi[i], (1 - 1.5 * (obj/phi[i]) +
                    0.5 * (obj/phi[i])^3), 0),
                  circular = {
                    obj.sc <- obj/phi[i];
                    obj.sc[obj.sc > 1] <- 1;
                    ifelse(obj < phi[i], (1 - (2 * ((obj.sc) *
                                                    sqrt(1 - ((obj.sc)^2)) +
                                                    asin(obj.sc)))/pi), 0)
                  },
                  cubic = {
                    obj.sc <- obj/phi[i];
                    ifelse(obj < phi[i], (1 - (7 * (obj.sc^2) -
                                               8.75 * (obj.sc^3) +
                                               3.5 * (obj.sc^5) -
                                               0.75 * (obj.sc^7))), 0)
                  },
                  powered.exponential = {
                    if(kappa > 2 | kappa <= 0)
                      stop("for power model parameter phi must be in (0,2]"
                           );
                    exp( - ((obj/phi[i])^kappa))
                  },
                  cauchy = (1 + (obj/phi[i])^2)^(-1 * kappa),
                  gneiting = {
                    obj.sc <- obj/phi[i];
                    t2 <- (1 - obj.sc);
                    t2 <- ifelse(t2 > 0, (t2^8), 0);
                    (1 + 8 * obj.sc + 25 * (obj.sc^2) + 32 * (obj.sc^
                                                              3)) * t2
                  },
                  gneiting.matern = { 
                    obj.sc <- obj/(phi[i] * kappa[2]);
                    t2 <- (1 - obj.sc);
                    t2 <- ifelse(t2 > 0, (t2^8), 0);
                    cov <- (1 + 8 * obj.sc + 25 * (obj.sc^2) + 32 * (obj.sc^3)) * t2;
                    cov * matern(u = obj, phi = phi[i], kappa = kappa[1])
                  },
                  stop("wrong or no sppecification of cov.model")
                  )
    cov <- sigmasq[i] * cov
    covs <- covs + cov
  }
  return(covs)
}


"dist0.krige" <-
function (x0, coords) 
{
  if (length(x0) != 2) 
    stop(paste("length of x0 is", length(x0), "(it must be 2)"))
  coords[, 1] <- coords[, 1] - x0[1]
  coords[, 2] <- coords[, 2] - x0[2]
  return(sqrt(coords[, 1]^2 + coords[, 2]^2))
}
"dist2full" <-
  function (dis) 
{
  n <- attr(dis, "Size")
  full <- matrix(0, n, n)
  full[lower.tri(full)] <- dis
  full + t(full)
}






