"krige.bayes" <- 
function(geodata, coords=geodata$coords, data=geodata$data, locations = "no",
         trend.d = "cte", trend.l = "cte",
	beta.prior = "flat", beta = NULL, beta.var = NULL, sill.prior = TRUE,
         sill = NULL, 
         range = NULL, range.prior = c("uniform", "exponential", "fixed", "squared.reciprocal",
	"reciprocal"), exponential.prior.par = 1,
	range.limits = "default", n.range = 51, cov.model = c("exponential", "matern", 
	"gaussian", "spherical", "circular", "cubic", "wave", "powered.exponential", 
	"cauchy", "gneiting", "gneiting.matern", "pure.nugget"), kappa = 0.5,
	nugget.prior = c("fixed", "uniform"), nugget = 0, lambda = 1, 
	aniso.pars = FALSE,  n.posterior = 1000, distribution = TRUE,
	mean.estimator = TRUE, quantile.estimator = FALSE, probability.estimator = FALSE,
	messages.screen = TRUE, signal = FALSE, info.for.prediction = TRUE, moments = TRUE)
{
	# This function computes (Bayesian) estimates of a spatial linear model and performs Bayesian and/or kriging prediction in a set of locations specified by the user.
	# Priors options for mean and/or covariance parameters 
	# coords       : data coordinates (vector for 1d or matrix for 2d data)
	# data         : vector with data values
	# locations    : coordinates of points to be estimated (vector for 1d or matrix for 2d data). If locations='no' only model parameters estimates are returned if the full bayesian model is considered.
	# trend.d      : trend data in data locations. Default is constant trend. The options '1st' or '2nd' builts a first or second degree polinomial.
	# trend.l      : trend data in locations to be estimated. Default is constant trend. The options '1st' or '2nd' builts a first or second degree polinomial.
	# beta.prior   : 'deg', 'flat', 'normal'
	##             'deg' (degenerated) implies in simple kriging and a value of beta must be provided
	##             'flat' (locally uniform or Jeffrey's) implies in ordinary kriging. No value of beta is needed
	##             'normal' prior parameters must be provided in beta and beta.var
	# sigma.prior  : prior for sigmasq, if T prior=1/sigmasq ; if F sigmasq is consider as a known parameter (TO IMPLEMENT: conjugate prior)
	# beta         : prior mean for parameter beta (according to beta.prior option)
	# beta.var     : prior variance (or covariance matrix) for parameter beta
	# sill         : sill parameter if no prior for sill
	# sill.prior   : prior for sill (T or F)
	# range        : range parameter (if no prior for phi)
	# range.prior  : prior for phi. Options for 'reciprocal' (prior=1/phi), squared.reciprocal (prior=1/phi^2) 'uniform' (prior \propto 1), exponential (exp(-c*phi)), or F to consider phi as a known parameter
	# exponential.prior.par: constante parameter for the exponential prior for phi(default=1)
	# range.limits : limits for distribution of range. Only if range.prior=T. Default is half of maximum distance between 2 data points within the area
	# cov.model    : 'matern'(requires values for kappa), 'sph', 'gauss', 'wave', 'power', 'powered.exponential'(requires values for kappa), cauchy(requires values for kappa), gneiting and gneiting- matern(requires values for kappa) 
	## for 'power' model cov = cte - sigmasq * h^phi 
	# kappa        : A number or vector required only if cov.model='matern' or 'powered.exponential'
	# nugget       : nugget effect (tausq). Allowed only if sill.prior=range.prior=F
	# aniso.pars   : anisotropy parameters (rotation, ranges ratio)
	# n.range      : number of points for the range's discrete distribution. Only if range.prior=T
	# n.posterior  : number of points for the location's discrete distributions. Only if range.prior=T
	# distribution : If T the function returns the (sampled) posterior for each location
	# mean.estimator: If true objects "predict.mean" and krige.var are returned.
	# quantile.estimator: the default is F. If a vector with numbers in the interval $[0,1]$ is provided the corresponding quantile(s) estimators are returned in "$quantiles". For example, if estimator=c(0.25,0.50,0.75) the function returns the 1st quartile, median and 3rd quartile for each location. If T, the default is the vector c(0.025, 0.5, 0.975) and an useful uncertainty measure which is an analogy to the kriging standard error can be computed by $(q0.975 - q0.025)/4$.
	# probability.estimator. The default is F. If some cutoff values are providedinstead, an object called "$probability" is returned for each location, with the probabilities of the variable be less or equal than the values provided in the argument.
	# signal: if true the signal is predict, otherwise the variable is predict. The expectations results are the same and the difference is only in the kriging variance, if nugget fifferent form zero 
	# nugget: can be a fixed value (default is zero) or, for a full Bayesian approach  a vetor with relative values (tausq/sigmasq) which will be use to build the discrete grid with the range (phi).
	# lambda : the Box-Cox transformation parameter.If $\lambda=1$ no transformation is performed otherwise the data is transformed and backtransformed within the code. The results are in the same scale as the original data. If $\lambda=0$ the log-transformation is performed. If $\lambda < 0$ the mean predictor doesn't make sense (the resulting distribution has no expectation). This trendsformation is available only for the "full Bayesian approach.
	# info.for.prediction: if T intermediate results are kept to be used in future predictions (to be done using the function krige.bayes.predict)
  require(mva)
  call <- match.call()
  cov.model <- match.arg(cov.model)
  range.prior <- match.arg(range.prior)
  nugget.prior <- match.arg(nugget.prior)
  if(nugget.prior == "fixed")
    nugget.prior <- FALSE
  seed <- .Random.seed
        #
	#########################PART 1##############################
	# Checking the input
	#
  data.dist <- as.vector(dist(coords))
	# check == here
  if(min(data.dist) == 0) stop(
          "This function does not allows for two data at same location")
  if(all(locations == "no")) {
    if(beta.prior == "flat" & sill.prior == TRUE & range.prior != "fixed")
      print("No prediction locations provided.  Only Bayesian estimates of model parameters will be returned"
            )
    else stop("Locations to be predicted not provided. Bayesian estimation allowed only for the full Bayesian (uncertainty) model"
              )
  }
  if(length(sill) > 1 | length(range) > 1)
    stop("Bayesian kriging not (yet) implemented for nested variogram models"
         )
  if(nugget.prior == FALSE && length(nugget) > 1)
    stop("to include the nugget in the Bayesian analysis set the argument nugget.prior"
         )
  if((length(nugget) > 1) & any(nugget > 0.5))
    print("WARNING: when nugget is considered random in the Bayesian analysis the values assigned in the argment nugget must be relative nuggets: tausq/sigmasq"
          )
  if(sill.prior == FALSE & is.numeric(sill) == FALSE)
    stop("sill parameter needed if no prior for sill")
  if(beta.prior != "flat" & sill.prior == TRUE)
    stop("selected prior for beta not allowed for selected prior for sill"
         )
  if(range.prior == "fixed" & is.numeric(range) == FALSE)
    stop("range parameter needed if no prior for range")
  if(any(aniso.pars) != FALSE)
    warning("anisotropy parameters assumed to be constants")
  if(is.numeric(quantile.estimator))
    if(any(quantile.estimator) < 0 | any(quantile.estimator) > 1)
      stop("quantiles indicators must be numbers in the interval [0,1]"
           )
  if(quantile.estimator == TRUE)
    quantile.estimator <- c(0.025000000000000001, 0.5, 
                            0.97499999999999998)
  if(probability.estimator == TRUE)
    stop("For probability.estimation != FALSE a numeric cutoff value (or vector) must be provided in this argument"
         )
  if(lambda != 1 & lambda != 0 & moments == TRUE) {
    moments <- FALSE
    cat(paste("WARNING: moments cannot be computed for lambda =", 
                                                              lambda, ". Argument moments was set to FALSE\n"))
  }
  if(lambda < 0 & mean.estimator == TRUE) {
    mean.estimator <- FALSE
    print("mean.predictor set to F. The resulting distribution does not has expectation for $lambda < 0$"
          )
  }
                                        #
                                        # Transforming (Box-Cox) the data (if the case)
                                        #
  if(lambda != 1) {
    if(range.prior == "fixed")
      stop("transformation option available only for the full Bayesian approach"
           )
    if(messages.screen == TRUE)
      print(paste(
                  "Data transformation (Box-Cox) for $lambda=",
                  lambda), seq = "")
    if(lambda == 0)
      data <- log(data)
    else data <- ((data^lambda) - 1)/lambda
  }
                                        #
                                        # Checking the dimension of points to be predict (1 or 2)	
                                        #
  if(is.vector(coords)) coords <- as.matrix(cbind(coords, 0)) else coords <-
    as.matrix(coords)
  if(all(locations != "no")) {
    if(is.vector(locations)) {
      if(length(locations) == 2) {
        locations <- t(as.matrix(locations))
        print("WARNING: THE FUNCTION IS CONSIDERING THAT YOU HAVE ENTERED WITH 1 POINT TO BE PREDICTED IN A TWO DIMENSION REGION"
              )
      }
      else locations <- as.matrix(cbind(locations, 0))
    }
    else locations <- as.matrix(locations)
  }
                                        #
                                        # Trend matrices:	
                                        #
  if(inherits(trend.d, "formula") | inherits(trend.l, "formula")){
    if(any(locations != "no"))
      if((inherits(trend.d, "formula") == FALSE) | (inherits(trend.l, "formula") == FALSE))
        stop("trend.d and trend.l must have similar specification")
    if(messages.screen == TRUE)
      print("Kriging with external trend to be performed using covariates provided by the user")
  }
  else{
    if (any(locations != "no") & (trend.d != trend.l)){
      stop("trend.l is different from trend.d")
    }
    if(messages.screen == TRUE){
      if(trend.d == "cte")
        print("analysis assuming constant mean")
      if(trend.d == "1st")
        print("analysis assuming a 1st degree polinomial trend")
      if(trend.d == "2nd") 
        print("analysis assuming a 2nd degree polinomial trend")
    }
  }
  trend.d <- trend.spatial(trend=trend.d, coords=coords)
  trend.l.pred <-  trend.l
  if(any(locations != "no"))
    trend.l <- trend.spatial(trend=trend.l, coords=locations)
                                        #
                                        # Eliminating row and columns names
                                        #
  dimnames(coords) <- list(NULL, NULL)
  dimnames(trend.d) <- list(NULL, NULL)
  if(all(locations != "no")) {
    dimnames(locations) <- list(NULL, NULL)
    dimnames(trend.l) <- list(NULL, NULL)
  }
                                        #
                                        # Checking dimensions
                                        #
  if(nrow(coords) != length(data)) stop(
           "number of data is different of number of data locations (coordinates)"
           )
  if(all(locations != "no")) {
    locations <- as.matrix(locations)
    if(nrow(locations) != nrow(trend.l))
      stop("number of points to be estimated is different of the number of trend points"
           )
    dimnames(locations) <- list(NULL, NULL)
  }
  dimnames(coords) <- list(NULL, NULL)
                                        #
#########################PART 3##############################
                                        # Anisotropy correction
                                        #
  if(any(aniso.pars) == TRUE) {
    coords <- coords.aniso(coords = coords, aniso.pars = aniso.pars
                           )
    if(all(locations != "no"))
      locations <- coords.aniso(coords = locations, 
				aniso.pars = aniso.pars)
  }
                                        #
#########################PART 4##############################
                                        # Preliminary kriging matrices calculations
                                        #
  n <- length(data)
  if(is.matrix(trend.l))
    ni <- dim(trend.l)[1]
  else ni <- length(trend.l)
  tausq <- nugget
  if(sill.prior == TRUE)
    sigmasq <- 1
  else sigmasq <- sill
  if(range.prior == "fixed") {
    phi <- range
                                        # Computing covariance matrix for data points
    invcov <- varcov.spatial(coords = coords, cov.model = cov.model,
                             kappa = kappa, nugget = nugget, cov.pars = c(sigmasq,
                                                               phi), inv = TRUE, scaled = TRUE)
                                        #	
                                        # Computing covariance vector between data points and locations to be estimated
                                        #	
    dist0mat <- apply(locations, 1, dist0.krige, coords = coords)
    v0mat <- cov.spatial(obj = dist0mat, cov.model = cov.model,
                         kappa = kappa, cov.pars = c(1, phi))
                                        #	
                                        # Some matrices computations
                                        #
    ttiv <- crossprod(trend.d, invcov)
    ttivtt <- ttiv %*% trend.d
    ittivtt <- solve(ttivtt)
    ttivz <- ttiv %*% data
    tv0iv <- t(apply(v0mat, 2, crossprod, y = invcov))
    tv0ivz <- tv0iv %*% data
    tv0ivx <- apply(tv0iv, 1, crossprod, y = trend.d)
    tv0ivv0 <- diag(tv0iv %*% v0mat)
    tb <- t(trend.l) - tv0ivx
    b <- t(tb)
                                        #
#########################PART 5##############################
                                        # Simple kriging
                                        #
    if(beta.prior == "deg") {
      beta.mean <- paste("SK : beta provided by user: ",
                         beta)
      predict <- tv0ivz + b %*% beta
      if(signal == TRUE)
        krige.var <- sigmasq * (1 - tv0ivv0)
      else krige.var <- nugget + sigmasq * (1 - tv0ivv0)
      priors.messa <- c("no priors for parameters", 
                        "results corresponds to simple kriging")
      sill.mean <- paste("sill provided by user: ", sill)
    }
                                        #
#########################PART 6##################################
                                        # Ordinary kriging (uncertainty only in the mean with flat prior)
                                        #
    if(beta.prior == "flat") {
      beta.mean <- ittivtt %*% ttivz
      predict <- tv0ivz + b %*% beta.mean
      bi <- apply(tb, 2, crossprod, y = ittivtt)
      if(ncol(trend.d) > 1)
        bi <- t(bi)
      bitb <- diag(bi %*% tb)
      if(sill.prior == FALSE) {
        if(signal == TRUE)
          krige.var <- sigmasq * (1 - tv0ivv0 +
                                  bitb)
        else krige.var <- nugget + sigmasq * (1 - 
                                              tv0ivv0 + bitb)
        priors.messa <- c(
                          "uninformative prior for beta", 
                          "no prior for sill", 
                          "no prior for range", 
                          "kriging estimates and variances equivalent to ordinary kriging"
                          )
        sill.mean <- paste("sill provided by user: ",
                           sill)
      }
      else {
        resid <- as.vector(data - trend.d %*% beta.mean
                           )
        sill.mean <- (resid %*% invcov %*% resid)/
          (n - length(beta.mean) - 2)
        if(signal == TRUE)
          krige.var <- sill.mean * (1 - tv0ivv0 +
                                    bitb)
        else krige.var <- nugget + sill.mean * (1 -
						tv0ivv0 + bitb)
        priors.messa <- c(
                          "uninformative prior for beta", 
                          "uninformative prior for sill", 
                          "kriging estimates (but not variances) equivalent to ordinary kriging"
                          )
      }
    }
                                        #
#########################PART 8##############################
                                        # Uncertainty only in the mean parameter (with normal prior)
                                        #
    if(beta.prior == "normal") {
      ibeta.var <- solve(beta.var)
      varbgz <- solve(ibeta.var + ttivtt/sigmasq)
      beta.mean <- varbgz %*% (beta %*% ibeta.var + ttivz/
                               sigmasq)
      predict <- tv0ivz + b %*% beta.mean
      bi <- apply(tb, 2, crossprod, y = varbgz)
      bitb <- diag(bi %*% tb)
      if(signal == TRUE)
        krige.var <- sigmasq * (1 - tv0ivv0) + bitb
      else krige.var <- nugget + sigmasq * (1 - tv0ivv0) +
        bitb
      priors.messa <- c("Normal prior for beta", 
                        "no prior for sill", "no prior for range")
      sill.mean <- paste("provided by user: ", sill)
    }
                                        #
                                        # Preparing output
                                        #
    results <- list(predict = as.vector(predict), krige.var = 
                    as.vector(krige.var), beta.mean = beta.mean, 
                    sigmasq.mean = sill.mean, paste(
                      "range provided by user: ", range), type.kriging = 
                    priors.messa)
  }
  else {
    if(messages.screen == TRUE)
      print("Computing the discrete approximation for the posterior distribution of phi"
            )
                                        #	
#########################PART 9.1##############################
                                        # Computing a discrete probability distribution for phi
                                        #
    beta.size <- dim(trend.d)[2]
    gl <- as.vector(n - beta.size)
    n.nugget <- length(nugget)
    phidist <- list()
    if(all(range.limits == "default"))
      phidist$phi <- seq(0, max(data.dist), l = n.range)
    else {
      if(length(range.limits) == 2)
        phidist$phi <- seq(range.limits[1], 
                           range.limits[2], l = n.range)
      else {
        phidist$phi <- range.limits
        n.range <- length(phidist$phi)
      }
    }
#    if(range.prior == "reciprocal" | range.prior == 
#       "squared.reciprocal") {
#                                        # check == here
#      if(phidist$phi[1] == 0) phidist$phi[1] <- 
#        0.10000000000000001 * (phidist$phi[
#                                           2] - phidist$phi[1])
#    }
    phi.val <- phidist$phi
    phidist$phi <- as.matrix(expand.grid(phidist$phi, nugget))
    dimnames(phidist$phi) <- list(NULL, NULL)
    temp.res <- apply(phidist$phi, 1, krige.bayes.aux1, coords = 
                      coords, trend.d = trend.d, data = data, cov.model = 
                      cov.model, kappa = kappa, range.prior = range.prior,
                      exponential.prior.par = exponential.prior.par, n = n)
    extract.f <- function(obj)
      {
        return(obj$s2phi)
      }
    phidist$s2 <- matrix(sapply(temp.res, extract.f), nrow = 
                         n.range, ncol = n.nugget)
    extract.f <- function(obj)
      {
        return(obj$logprobphi)
      }
    phidist$logprobphi <- matrix(sapply(temp.res, extract.f), nrow = 
                              n.range, ncol = n.nugget)
    extract.f <- function(obj)
      {
        return(obj$inv.lower)
      }
    phidist$inv.lower <- array(t(sapply(temp.res, extract.f)),
                               dim = c(n.range, n.nugget, (n * (n - 1))/2))
    extract.f <- function(obj)
      {
        return(obj$inv.diag)
      }
    phidist$inv.diag <- array(t(sapply(temp.res, extract.f)), dim
                              = c(n.range, n.nugget, n))
    extract.f <- function(obj)
      {
        return(obj$vbetaphi)
      }
    dim.trend <- dim(trend.d)[2]
    if(dim.trend == 1)
      phidist$vbeta.vector <- array(matrix(sapply(temp.res,
                                                  extract.f), nrow = n.range, ncol = n.nugget),
                                    dim = c(n.range, n.nugget, 1))
    else phidist$vbeta.vector <- array(t(sapply(temp.res, extract.f
                                                )), dim = c(n.range, n.nugget, (dim.trend^
                                                      2)))
    extract.f <- function(obj)
      {
        return(obj$betaphi)
      }
    if(dim.trend == 1)
      phidist$beta <- array(matrix(sapply(temp.res, extract.f
                                          ), nrow = n.range, ncol = n.nugget), dim = c(
                                                                                 n.range, n.nugget, 1))
    else phidist$beta <- array(t(sapply(temp.res, extract.f)),
                               dim = c(n.range, n.nugget, (dim.trend)))
    temp.res <- NULL
                                        #    
    phidist$logprobphi <- phidist$logprobphi + abs(min(phidist$
                                                       logprobphi))
    phidist$probphi <- exp(phidist$logprobphi)
    sumprob <- sum(phidist$probphi)
    phidist$probphi <- phidist$probphi/sumprob
                                        #
#########################PART 9.2##############################
                                        # Sampling from the parameters distribution
                                        #
    if(messages.screen == TRUE) print(
         "Sampling from multivariate distribution of the parameters"
         )
    ind <- sample((1:(n.range * n.nugget)), n.posterior, replace = 
                  T, prob = as.vector(phidist$probphi))
    ind.unique <- sort(unique(ind))
    ind.length <- length(ind.unique)
    ind.table <- table(ind)
    phi.unique <- phidist$phi[ind.unique,  ]
    if(messages.screen == TRUE) {
      print("the sampled values from the phi/tau.rel distribution and their frequencies are:"
            )
      print(rbind(phi = phi.unique[, 1], nugget.rel = 
                  phi.unique[, 2], frequency = ind.table))
    }
    phi.sam <- phidist$phi[ind,  ]
    vecpars.back.order <- order(ind)
    vec.s2 <- rep(as.vector(phidist$s2)[ind.unique], ind.table)
    samples.chisq <- rchisq(n.posterior, df = gl)
    sigmasq <- (gl * vec.s2)/samples.chisq
    if(beta.size == 1) {
      samples.beta <- rnorm(n.posterior, mean = 0, sd = 1)
      vec.beta <- rep(as.vector(phidist$beta)[ind.unique],
                      ind.table)
      vec.vbeta <- rep(as.vector(phidist$vbeta.vector)[
                                                       ind.unique], ind.table)
      beta <- vec.beta + sqrt(sigmasq * vec.vbeta) * 
        samples.beta
    }
    else {
      ind.beta <- matrix(phidist$beta, ncol = beta.size)[
                                         ind.unique,  ]
      ind.beta <- ind.beta[rep(1:ind.length, ind.table),
                           ]
      ind.vbeta <- matrix(phidist$vbeta.vector, ncol = 
                          beta.size^2)[ind.unique,  ]
      ind.vbeta <- ind.vbeta[rep(1:ind.length, ind.table),
                             ] * sigmasq
                                        #      print("2.4: try to speed up thisbit!")
      temp.res <- apply(ind.vbeta, 1, krige.bayes.aux3, 
                        beta.size = beta.size)
      beta <- ind.beta + t(temp.res)
      temp.res <- NULL
    }
    if(beta.size == 1) {
      trend.mean <- mean(beta)
      trend.median <- median(beta)
    }
    else {
      trend.mean <- apply(beta, 2, mean)
      trend.median <- apply(beta, 2, median)
    }
    sill.mean <- mean(sigmasq)
    sill.median <- median(sigmasq)
    range.marg <- apply(phidist$probphi, 1, sum)
    range.marg <- range.marg/(sum(range.marg))
    range.mean <- phi.val %*% range.marg
    range.median <- median(phi.sam[, 1])
                                        # check == here
    range.mode <- phi.val[range.marg == max(range.marg)]
    nugget.marg <- apply(phidist$probphi, 2, sum)
    nugget.marg <- nugget.marg/(sum(nugget.marg))
    nugget.mean <- nugget %*% nugget.marg
    nugget.median <- median(phi.sam[, 2])
                                        # check == here
    nugget.mode <- nugget[nugget.marg == max(nugget.marg)]
                                        #
                                        # Computing the conditional (on phi nad tausq.rel if the case) modes for beta and sigmasq
                                        #
    invcov.mode <- varcov.spatial(coords = coords, cov.model = 
                                  cov.model, kappa = kappa, nugget = nugget.mode, 
                                  cov.pars = c(1, range.mode), inv = TRUE)
    ivt <- crossprod(invcov.mode, trend.d)
    beta.mode.cond <- solve(crossprod(ivt, trend.d), crossprod(
                                                               ivt, data))
    resid.mode <- as.vector(data - trend.d %*% beta.mode.cond)
    sill.mode.cond <- (resid.mode %*% invcov.mode %*% resid.mode)/
      (n - length(beta.mode.cond) + 2)
    if(beta.size == 1)
      beta.summary <- c(mean = trend.mean, median = 
                        trend.median, mode.cond = beta.mode.cond)
    else beta.summary <- cbind(mean = trend.mean, median = 
                               trend.median, mode.cond = beta.mode.cond)
    if(all(locations != "no") | info.for.prediction == TRUE) {
      pred.list <- list(coords = coords, cov.model = 
                        cov.model, kappa = kappa, trend.d = trend.d,
                        data = data, beta.size = beta.size, n = n,
                        phidist = phidist, phi.val = phi.val, nugget = 
                        nugget, distribution = distribution, 
                        n.posterior = n.posterior, ind.length = 
                        ind.length, vecpars.back.order = 
                        vecpars.back.order, phi.unique = phi.unique,
                        ind.table = ind.table, gl = gl, mean.estimator
                        = mean.estimator, quantile.estimator = 
                        quantile.estimator, probability.estimator = 
                        probability.estimator, ind = ind, lambda = 
                        lambda, moments = moments)
    }
    if(all(locations != "no")) {
      if(messages.screen == TRUE)
        print("Prediction at selected locations")
                                        #	
                                        # sampled phi from previous section
                                        #
                                        #			print(.Random.seed)
      results <- krige.bayes.predict(locations = locations,
                                     trend.l = trend.l.pred, signal = signal, 
                                     info = pred.list, messages.screen = 
                                     messages.screen, save.call = FALSE)
      results$beta.summary <- beta.summary
      results$sigmasq.summary <- c(mean = sill.mean, median
                                   = sill.median, mode.cond = sill.mode.cond)
      results$phi.summary <- c(mean = range.mean, median = 
                               range.median, mode = range.mode)
      if(nugget.prior != FALSE)
        results$tausq.summary <- c(mean = nugget.mean,
                                   median = nugget.median, mode = 
                                   nugget.mode)
      results$beta.samples <- as.matrix(beta)[
                                              vecpars.back.order,  ]
      if(beta.size == 1)
        results$beta.samples <- as.vector(results$
                                          beta.samples)
      results$sigmasq.samples <- sigmasq[vecpars.back.order]
      results$phi.samples <- phi.sam[vecpars.back.order,
                                     1]
      results$tausq.samples <- phi.sam[vecpars.back.order,
                                       2]
    }
    else {
      results <- list(predict = 
                      "no locations to be predicted", krige.var = 
                      "no locations to be predicted", beta.summary = 
                      beta.summary, sigmasq.summary = c(mean = 
                                      sill.mean, median = sill.median, mode.cond = 
                                      sill.mode.cond), phi.summary = c(mean = 
                                                         range.mean, median = range.median, mode = 
                                                         range.mode), tausq.summary = c(mean = 
                                                                        nugget.mean, median = nugget.median, mode = 
                                                                        nugget.mode), beta.samples = as.matrix(beta)[
                                                                                        vecpars.back.order,  ], sigmasq.samples = 
                      sigmasq[vecpars.back.order], phi.samples = 
                      phi.sam[vecpars.back.order, 1], tausq.samples
                      = phi.sam[vecpars.back.order, 2])
      if(beta.size == 1)
        results$beta.samples <- as.vector(results$
                                          beta.samples)
      results$message <- 
        "Only Bayesian estimation of model parameters were computed"
      if(messages.screen == TRUE)
        print(results$message)
    }
  }
  if(all(nugget != 0) & length(nugget) == 1)
    results$nugget.fixed <- nugget
  results$.Random.seed <- seed
  if(messages.screen == TRUE)
    print("done!")
  if(info.for.prediction == TRUE & range.prior != "fixed")
    results$info.for.prediction <- pred.list
  results$call <- call
  class(results) <- c("krige.bayes", "kriging")
  return(results)
}

"krige.bayes.aux1" <- 
function(phinug, coords, trend.d, data, cov.model, kappa, range.prior, 
	exponential.prior.par, n)
{
	# auxiliary function to perform calculation for the function krige.bayes
	phi <- phinug[1]
	nugget <- phinug[2]
	if(phi == 0)
          covphi <- list(inverse = diag((1/(1 + nugget)), n), 
			log.det.to.half = ((n/2) * log(1 + nugget)))
        else 
          covphi <- varcov.spatial(coords = coords, cov.model = 
                                   cov.model, kappa = kappa, nugget = nugget,
                                   cov.pars = c(
                                     1, phi), inv = TRUE, det = TRUE)
	ttiv <- crossprod(trend.d, covphi$inverse)
	ttivtt <- ttiv %*% trend.d
	vbetaphi <- solve(ttivtt)
	eival.x <- eigen(ttivtt, symmetric = TRUE, only.values = TRUE)
	ttivttdet <- prod(eival.x$values)
	betaphi <- as.vector(vbetaphi %*% ttiv %*% data)
	res <- data - trend.d %*% betaphi
	sqres <- crossprod(res, covphi$inverse) %*% res
	gl <- length(data) - length(betaphi)
	s2phi <- (1/gl) * sqres
	logprobphi <- (-0.5) * log(ttivttdet) - (covphi$log.det.to.half) - (
		gl/2) * log(s2phi)
	if(range.prior == "reciprocal" & round(1e+08 * phi) != 0)
		logprobphi <- logprobphi - log(phi)
	if(range.prior == "squared.reciprocal" & round(1e+08 * phi) != 0)
		logprobphi <- logprobphi - log(phi^2)
	if(range.prior == "exponential")
		logprobphi <- logprobphi - exponential.prior.par * phi
#	probphi <- (ttivttdet^(-0.5) * 1)/exp(covphi$log.det.to.half) * s2phi^
#		( - gl/2)
#	if(range.prior == "reciprocal")
#		probphi <- probphi/phi
#	if(range.prior == "squared.reciprocal")
#		probphi <- probphi/(phi^2)
#	if(range.prior == "exponential")
#		probphi <- probphi * exp( - exponential.prior.par * phi)
	inv.lower <- covphi$inverse[lower.tri(covphi$inverse)]
	inv.diag <- diag(covphi$inverse)
	results <- list(vbetaphi = vbetaphi, betaphi = betaphi, s2phi = s2phi,
		logprobphi = logprobphi, inv.lower = inv.lower, inv.diag = inv.diag)
	return(results)
}

"krige.bayes.aux10" <- 
function(phinug, d0mat, ind.coincide, data.coincide, pred.list, ni, trend.l,
	signal, messages.screen)
{
	n <- pred.list$n
	phinug <- as.vector(phinug)
	phi.ind <- order(pred.list$phi.val)[round(100000000. * pred.list$
		phi.val) == round(100000000. * phinug[1])]
	nug.ind <- order(pred.list$nugget)[round(100000000. * pred.list$nugget) ==
		round(100000000. * phinug[2])]
	iv <- matrix(0, n, n)
	iv[lower.tri(iv)] <- as.vector(pred.list$phidist$inv.lower[phi.ind, nug.ind,
		])
	iv <- iv + t(iv)
	diag(iv) <- as.vector(pred.list$phidist$inv.diag[phi.ind, nug.ind,  ])
	v0 <- cov.spatial(obj = d0mat, cov.model = pred.list$cov.model, kappa
		 = pred.list$kappa, cov.pars = c(1, phinug[1]))
	ivv0 <- iv %*% v0
	iv <- NULL
	tb <- trend.l - crossprod(ivv0, pred.list$trend.d)
	tmean <- crossprod(ivv0, pred.list$data) + tb %*% as.vector(pred.list$phidist$
		beta[phi.ind, nug.ind,  ])
	if(round(100000000. * phinug[2]) == 0)
		tmean[ind.coincide] <- pred.list$data[data.coincide]
	s2i <- pred.list$phidist$s2[phi.ind, nug.ind]
	vbetai <- matrix(as.vector(pred.list$phidist$vbeta.vector[phi.ind, nug.ind,  ]),
		ncol = pred.list$beta.size, nrow = pred.list$beta.size)
	tv0ivv0 <- apply(as.matrix(1:ni), 1, krige.bayes.aux11, mat1 = v0,
		mat2 = ivv0, by = "col")
	tbivb <- tb %*% vbetai
	tbivbb <- apply(as.matrix(1:ni), 1, krige.bayes.aux11, mat1 = tb, mat2
		 = tbivb, by = "row")
	if(signal == TRUE)
		tvar <- s2i * (1 - tv0ivv0 + tbivbb)
	else tvar <- s2i * ((1 + phinug[2]) - tv0ivv0 + tbivbb)
	if(round(100000000. * phinug[2]) == 0)
		tvar[ind.coincide] <- 0
	tv0ivv0 <- NULL
	tbivbb <- NULL
	tbivb <- NULL
	if(messages.screen == TRUE) {
		if(pred.list$ind.length <= 20)
			print(paste("prediction for the parameter set", 
				.temp.ap, "out of", pred.list$ind.length))
		if(pred.list$ind.length > 20 & pred.list$ind.length <= 200)
			if(.temp.ap %% 10 == 1)
				print(paste("prediction for the parameter set",
					.temp.ap, "out of", pred.list$
					ind.length))
		if(pred.list$ind.length > 200)
			if(.temp.ap %% 100 == 1)
				print(paste("prediction for the parameter set",
					.temp.ap, "out of", pred.list$
					ind.length))
	}
	assign(".temp.ap", (.temp.ap + 1), pos = 1)
	return(cbind(tmean, tvar))
}

"krige.bayes.aux11" <- 
#
function(i, mat1, mat2, by)
{
	if(by == "row")
		return(mat1[i,  ] %*% mat2[i,  ])
	if(by == "col")
		return(mat1[, i] %*% mat2[, i])
}

"krige.bayes.aux12" <- 
function(phinug, d0mat, ind.coincide, data.coincide, pred.list, ni, trend.l,
	signal, messages.screen, n.disc)
{
	n <- pred.list$n
	phinug <- as.vector(phinug)
	phi.ind <- order(pred.list$phi.val)[round(100000000. * pred.list$
		phi.val) == round(100000000. * phinug[1])]
	nug.ind <- order(pred.list$nugget)[round(100000000. * pred.list$nugget) ==
		round(100000000. * phinug[2])]
	iv <- matrix(0, n, n)
	iv[lower.tri(iv)] <- as.vector(pred.list$phidist$inv.lower[phi.ind, nug.ind,
		])
	iv <- iv + t(iv)
	diag(iv) <- as.vector(pred.list$phidist$inv.diag[phi.ind, nug.ind,  ])
	v0 <- cov.spatial(obj = d0mat, cov.model = pred.list$cov.model, kappa
		 = pred.list$kappa, cov.pars = c(1, phinug[1]))
	ivv0 <- iv %*% v0
	iv <- NULL
	tb <- trend.l - crossprod(ivv0, pred.list$trend.d)
	expec.y0 <- crossprod(ivv0, pred.list$data) + tb %*% as.vector(pred.list$phidist$
		beta[phi.ind, nug.ind,  ])
	if(round(100000000. * phinug[2]) == 0)
		expec.y0[ind.coincide] <- pred.list$data[data.coincide]
	s2i <- pred.list$phidist$s2[phi.ind, nug.ind]
	vbetai <- matrix(as.vector(pred.list$phidist$vbeta.vector[phi.ind, nug.ind,  ]),
		ncol = pred.list$beta.size, nrow = pred.list$beta.size)
	tv0ivv0 <- apply(as.matrix(1:ni), 1, krige.bayes.aux11, mat1 = v0,
		mat2 = ivv0, by = "col")
	tbivb <- tb %*% vbetai
	tbivbb <- apply(as.matrix(1:ni), 1, krige.bayes.aux11, mat1 = tb, mat2
		 = tbivb, by = "row")
	if(signal == TRUE)
		tvar <- s2i * (1 - tv0ivv0 + tbivbb)
	else tvar <- s2i * ((1 + phinug[2]) - tv0ivv0 + tbivbb)
	tvar <- (pred.list$gl/(pred.list$gl - 2)) * tvar
	expec.y0.2 <- tvar + (expec.y0)^2
	if(round(100000000. * phinug[2]) == 0)
		tvar[ind.coincide] <- 0
	tv0ivv0 <- NULL
	tbivbb <- NULL
	tbivb <- NULL
	if(messages.screen == TRUE) {
		if(n.disc <= 20)
			print(paste("prediction for the parameter set", 
				.temp.ap1, "out of", n.disc))
		if(n.disc > 20 & n.disc <= 200)
			if(.temp.ap1 %% 10 == 1)
				print(paste("prediction for the parameter set",
					.temp.ap1, "out of", n.disc))
		if(n.disc > 200)
			if(.temp.ap1 %% 100 == 1)
				print(paste("prediction for the parameter set",
					.temp.ap1, "out of", n.disc))
	}
	assign(".temp.ap1", (.temp.ap1 + 1), pos = 1)
	return(cbind(expec.y0, expec.y0.2))
}

"krige.bayes.aux2" <- 
function(x, cutoff)
{
	# auxiliary function to perform calculation for the function krige.bayes
	ncut <- length(cutoff)
	lx <- length(x)
	if(ncut > 1) {
		result <- rep(0, ncut)
		for(i in 1:ncut)
			result[i] <- sum(x <= cutoff[i])/lx
	}
	else result <- sum(x <= cutoff)/lx
	return(result)
}

"krige.bayes.aux3" <- 
##
# This function produces a sample from  a multivariate normal distribution 
# mean is 0 and cov is a vector of length beta.size^2
function(cov, beta.size)
{
	#  dm <- length(mean)
	#  dc <- dim(cov)[1]
	#  if(dm != dc)
	#    stop("mean and cov must have compatible dimensions")
	cov <- matrix(cov, ncol = beta.size)
	cov.svd <- svd(cov)
	cov.decomp <- cov.svd$u %*% (t(cov.svd$u) * sqrt(cov.svd$d))
	zsim <- as.vector(cov.decomp %*% rnorm(beta.size))
	return(zsim)
}

"krige.bayes.aux5" <- 
#
function(phi.s, d0, cov.model, kappa, trend.l, trend.d, data, beta.size, n,
	l)
{
	iv <- matrix(0, n, n)
	# check == here
	iv[lower.tri(iv)] <- as.vector(temp.phidist$inv.lower[temp.phidist$
		phi == phi.s,  ])
	iv <- iv + t(iv)
	# check == here
	diag(iv) <- as.vector(temp.phidist$inv.diag[temp.phidist$phi == phi.s,
		])
	v0 <- cov.spatial(obj = d0, cov.model = cov.model, kappa = kappa, 
		cov.pars = c(1, phi.s))
	tv0 <- t(v0)
	tv0iv <- tv0 %*% iv
	term1 <- tv0iv %*% data
	tb <- as.matrix(trend.l[l,  ] - tv0iv %*% trend.d)
	ttb <- t(tb)
	# check == here
	tmean <- term1 + ttb %*% as.vector(temp.phidist$beta[temp.phidist$
		phi == phi.s,  ])
	# check == here
	s2i <- as.vector(temp.phidist$s2[temp.phidist$phi == phi.s])
	# check == here
	vbetai <- matrix(as.vector(temp.phidist$vbeta.vector[temp.phidist$
		phi == phi.s,  ]), ncol = beta.size, nrow = beta.size)
	tvar <- s2i * (1 - tv0iv %*% v0 + ttb %*% vbetai %*% tb)
	return(list(tmean = tmean, tvar = tvar))
}

"krige.bayes.aux6" <- 
#
function(phi.s, d0, cov.model, kappa, trend.l, trend.d, data, beta.size, n,
	l, inv.lower, inv.diag, beta, s2, vbeta.vector)
{
	iv <- matrix(0, n, n)
	iv[lower.tri(iv)] <- inv.lower
	iv <- iv + t(iv)
	diag(iv) <- inv.diag
	v0 <- cov.spatial(obj = d0, cov.model = cov.model, kappa = kappa, 
		cov.pars = c(1, phi.s))
	tv0 <- t(v0)
	tv0iv <- tv0 %*% iv
	term1 <- tv0iv %*% data
	tb <- as.matrix(trend.l[l,  ] - tv0iv %*% trend.d)
	ttb <- t(tb)
	tmean <- term1 + ttb %*% as.vector(beta)
	s2i <- s2
	vbetai <- matrix(vbeta.vector, ncol = beta.size, nrow = beta.size)
	tvar <- s2i * (1 - tv0iv %*% v0 + ttb %*% vbetai %*% tb)
	return(list(tmean = tmean, tvar = tvar))
}

"krige.bayes.predict" <- 
  function(locations, trend.l, info, signal, n.predictive = "n.posterior",
           messages.screen = TRUE, save.call = TRUE, moments, distribution.keep,
           mean.estimator, quantile.estimator, probability.estimator)
{
  require(mva)
  if(save.call == TRUE)
    call <- match.call()
  if(missing(distribution.keep)) distribution <- info$distribution
  if(missing(mean.estimator)) mean.estimator <- info$mean.estimator
  if(missing(quantile.estimator)) quantile.estimator <- info$quantile.estimator
  if(missing(probability.estimator)) probability.estimator <- info$probability.estimator
  n <- info$n
  locations <- as.matrix(locations)
  dimnames(locations) <- list(NULL, NULL)
  trend.l <- trend.spatial(trend=trend.l, coords=locations)
  if(messages.screen == TRUE){
    if(inherits(trend.l, "formula"))
      print("prediction to be performed including (external) covariate(s) provided by the user")
    else{
      if(length(trend.l) == 1){
        if(trend.l == "cte")
          print("prediction to be performed with constant mean")
        if(trend.l == "1st") 
          print("prediction to be performed with 1st degree polinomial trend")
        if(trend.l == "2nd") 
          print("prediction to be performed with 2nd degree polinomial")	
      }
    }
  }
  ni <- dim(trend.l)[1]
  if(n.predictive == "n.posterior") {
    n.predictive <- info$n.posterior
    phi.sam <- info$phidist$phi[info$ind,  ]
    if(messages.screen == TRUE)
      print("samples of parameters for the predictive are the same as for the posterior")
  }
  else {
    ind <- sample((1:(dim(info$phidist$phi)[1])), n.predictive,
                  replace = TRUE, prob = as.vector(info$phidist$probphi
                                 ))
    ind.unique <- sort(unique(ind))
    info$ind.length <- length(ind.unique)
    info$ind.table <- table(ind)
    info$phi.unique <- info$phidist$phi[ind.unique,  ]
    if(messages.screen == TRUE) {
      print("the sampled values from the phi/tau.rel distribution and their frequencies are:"
            )
      print(rbind(phi = info$phi.unique[, 1], nugget.rel
                  = info$phi.unique[, 2], frequency = 
                  info$ind.table))
    }
    phi.sam <- info$phidist$phi[ind,  ]
    info$vecpars.back.order <- order(ind)
  }
  d0mat <- apply(locations, 1, dist0.krige, info$coords)
  ind.coincide <- (apply(d0mat < 1e-10, 2, sum) == 1)
  ind.coincide <- (1:ni)[ind.coincide]
  data.coincide <- (apply(d0mat < 1e-10, 1, sum) == 1)
  assign(".temp.ap", 1, pos = 1)
  temp.res <- apply(info$phi.unique, 1, krige.bayes.aux10, d0mat = 
                    d0mat, ind.coincide = ind.coincide, data.coincide = 
                    data.coincide, ni = ni, trend.l = trend.l, signal = signal,
                    pred.list = info, messages.screen = messages.screen)
  remove(".temp.ap", pos = 1)
  tmean <- t(temp.res[1:ni,  ])
  tsdv <- sqrt(t(temp.res[ - (1:ni),  ]))
  remove(temp.res)
#  tmean <- tmean[rep(1:info$ind.length, info$ind.table),  ]
#  tsdv <- tsdv[rep(1:info$ind.length, info$ind.table),  ] 
  results.pred <- list()
#  results.pred$dist.pred <- tmean + (tsdv * matrix(rnorm((n.predictive * ni)), nrow = 
#                                      n.predictive, ncol = ni) * sqrt(info$gl))/sqrt(matrix(
#                                                      rchisq((n.predictive * ni), df = info$gl), nrow = 
#                                                      n.predictive, ncol = ni))
  results.pred$dist.pred <- tmean[rep(1:info$ind.length, info$ind.table),  ] +
    (tsdv[rep(1:info$ind.length, info$ind.table),  ] *
     matrix(rnorm((n.predictive * ni)), nrow = 
            n.predictive, ncol = ni) * sqrt(info$gl))/sqrt(matrix(
                            rchisq((n.predictive * ni), df = info$gl), nrow = 
                            n.predictive, ncol = ni))
                                        #
                                        # Preparing output for posterior distributions
                                        #
  if(messages.screen == TRUE) print("preparing output of predictive distribution")
                                        #
                                        # Back transforming
                                        #
  if(info$lambda != 1) {
    print("Data transformation (Box-Cox) performed. Results returned in the original scale"
          )
    if(info$lambda == 0)
      results.pred$dist.pred <- exp(results.pred$dist.pred)
    else {
      if(info$lambda > 0)
        results.pred$dist.pred[results.pred$dist.pred < -1/info$lambda] <-
          -1/info$lambda
      if(info$lambda < 0)
        results.pred$dist.pred[results.pred$dist.pred > -1/info$lambda] <-
          -1/info$lambda
      results.pred$dist.pred <- ((results.pred$dist.pred * info$lambda) + 1)^
      (1/info$lambda)
    }
  }
                                        #
  if(mean.estimator == TRUE) {
    results.pred$predict.mean <- as.vector(apply(results.pred$dist.pred, 2,
                                                 mean))
    results.pred$krige.var <- as.vector(apply(results.pred$dist.pred, 2, var))
    message.prediction <- 
      "predicted mean at each location in $predict and predicted variances in $krige.var"
  }
  if(all(quantile.estimator != FALSE)) {
    results.pred$quantiles <- apply(results.pred$dist.pred, 2, quantile, probs
                                    = quantile.estimator)
    if(length(quantile.estimator) > 1) {
      results.pred$quantiles <- t(results.pred$quantiles)
      qname <- rep(0, length(quantile.estimator))
      for(i in 1:length(quantile.estimator))
        qname[i] <- paste("q", 100 * quantile.estimator[
                                                        i], sep = "")
      dimnames(results.pred$quantiles) <- list(NULL, qname)
    }
    else {
      results.pred$quantiles <- as.vector(results.pred$
                                          quantiles)
    }
    message.prediction <- 
      "Predicted quantile(s) at each location returned in $quantiles"
  }
  if(probability.estimator != FALSE) {
    results.pred$probability <- apply(results.pred$dist.pred, 2, 
                                      krige.bayes.aux2, cutoff = probability.estimator)
    if(length(probability.estimator) > 1)
      results.pred$probability <- t(results.pred$probability)
    message.prediction <- 
      "Estimated probabilities of being less than the provided cutoff(s), at each location, returned in $probability"
  }
  if(distribution == TRUE)
    results.pred$dist.pred <- results.pred$dist.pred[info$
                                        vecpars.back.order,  ]
  else results.pred$dist.pred <- NULL
  results.pred$type.kriging <- c(
                                 "Spatial prediction performed taking into account uncertainty in trend (mean), sill and range parameters")
  results.pred$message.prediction <- message.prediction
  phi.lev <- unique(info$phidist$phi[, 1])
  results.pred$phi.marginal.prediction <- data.frame(phi = phi.lev, 
                                                     expected = apply(info$phidist$probphi, 1, sum), sampled = 
                                                     as.vector(table(factor(phi.sam[, 1], levels = phi.lev)))/
                                                     n.predictive)
  nug.lev <- unique(info$phidist$phi[, 2])
  results.pred$nugget.marginal.prediction <- data.frame(nugget = nug.lev,
                                                        expected = apply(info$phidist$probphi, 2, sum), sampled = 
                                                        as.vector(table(factor(phi.sam[, 2], levels = nug.lev)))/
                                                        n.predictive)
  results.pred$call <- call
                                        #
                                        # Estimating the moments
                                        #
  if(missing(moments)) moments <- info$moments
  if(moments == TRUE) {
    print("computing moments of the predictive distributions")
    assign(".temp.ap1", 1, pos = 1)
    n.disc <- dim(info$phidist$phi)[1]
    temp.res <- apply(info$phidist$phi, 1, krige.bayes.aux12,
                      d0mat = d0mat, ind.coincide = ind.coincide, 
                      data.coincide = data.coincide, ni = ni, trend.l = 
                      trend.l, signal = signal, pred.list = info, 
                      messages.screen = messages.screen, n.disc = n.disc)
    remove(".temp.ap1", pos = 1)
    n.disc <- length(info$phidist$probphi)
    temp.res <- temp.res %*% info$phidist$probphi
    expec.y0 <- as.vector(temp.res[1:ni])
    expec.y0.2 <- as.vector(temp.res[(ni + 1):(2 * ni)])
    var.y0 <- expec.y0.2 - (expec.y0)^2
    if(info$lambda == 0) {
      temp <- expec.y0
      expec.y0 <- exp(temp + 0.5 * (var.y0))
      var.y0 <- (exp(2 * temp + var.y0)) * (exp(var.y0) -
                                            1)
      expec.y0.2 <- var.y0 + (expec.y0)^2
    }
    results.pred$moments <- cbind(expect.y0 = expec.y0, expect.y0.2
                                  = expec.y0.2, var.y0 = var.y0)
  }
  return(results.pred)
}
